/* >15impmain.c  */
/*-----------------------------------
14 April 2014: tauconst & samplebeta commented out
               only KMax = priorpar[6] used in the priorpar array

24 Jan 2013: Twister RN generator used

 2 Jan 2013: SDProbofK added

29 Dec 2012: DataFileName added as argument

                Has Init.d Out9.d as input

                Has Out7.d Out8.d as output

                Also outpdf.d as output if OUTPUTPDF flag is set to 0

1 May   2012: edi.bat

 GRAPHICSON 0/1 switches graphics off and on in impmain.c tgraph.c

            no graphics in tdist.c  tsubs.c


 compilable with gcc or Watcom

 mki.bat  makes o object file with gcc
 lki.bat  makes  impmain.exe  with g77
                   (requires impmain.0 tdist.o tsubs.o [tgraph.o not needed]

 mkiall.bat makes impmain.obj tsubs.obj tdist.obj tgraph.obj with wcl386

 lkiall.bat links above to make impmain.exe                  with wcl386




29 April 2012: Changes made in simplifying the format of Out8.d

               The following is now output

                        numoISsamples
                        priorKMax
                        model

                        for each k:
                            the best fit means, taus, awts

                        for each k:
                            the ProbofKs


                        this is new:

                        for each k:
                            k
                            the IS ests of means, taus, awts


               Detailed IS sample data removed.

               This version is run from <Mixdat29 with MCMC.xls>

25 April 2012: This also plots the histograms of the sample of IS values
               obtained for each k, to show how concentrated they are,
               which should be the case if the Importance Sampling is
               working effectively. Note that these plots have a logarithmic
               x-axis.

               Also bubble sort routines replaced by faster Shell sort
               routines.

28 July 2011:  This is the version using simplified importance sampling,
               with adjustment of probabilities to account for sample values
               which fall outside the allowed weight range, i.e. which
               take account of the acceptance/rejection ratio.

WATCH POINT: Check to see if CandWt[bestjtry] = 1.0 always
             so that candidate distribution only has one component.
search with WATCHP


3 Feb 2005 Samples only from the ranked simplex 

1 Oct 2004 Does all models using mean and SD as the parameters p1 and p2.
Cleaned up version with 'flip' removed

Handles mixtures by EM/nelmead, so every third parameter is a weight
Note nelmead returns neg of function it minimizes, i.e.
 the lik or logposterior, so largest is best.

This version unlike previous versions has the distribution dependent
subroutines split off as tdist.c

This version only does single fits.
However made easy to add new models.

Run through option MODEL and add it in. 
   tsubs needs altering in just one position
   tgraph need not be altered
if density/cdf used for likelihood calculation.

-------------------------------------*/
#include <math.h>
#include <float.h>
#include <stdio.h>
#include <time.h>

#include "threep9.h"

#define GRAPHICSON 0



#if GRAPHICSON
   #include "gengraph.h"
#endif

// OR  ||

#define MAXNUMCOMPS 13 // allows for 12 components
#define MAXNUMINTVLS 510
#define BSIZE 210000
#define TWO 2
#define REMOVEFACTOR 100.0 //factor used for deleting outlying IS obsns WAS 3.0
                          // was 10.0
#define HWIDTH  600
#define VHEIGHT 400
#define YCOVER  1.0

#define OUTPUTPDF 0
#define OUT7 1


long hitail;
long model;
long componentmodel;
long outflag; /* indicates if output is printed */
long spacingsflag;   /* for choice of function to be maximised */
long show;
long numevals;
long cdfavail;
double leftendvalue;
double gamlog, gamlog1, gamlog2;
long numcomps;
long maxnumlevs;
long effort;

double * yobs;
double yminposdiff;
long n;
long akey, bkey;

char * filedatptr;
FILE * fp1;
FILE * fp7; // used for parameter histogram info
FILE * fp9;

// ========= This next block is the twister RN generator =============

/* A C-program for MT19937: Real number version([0,1)-interval) */
/* (1999/10/28)                                                 */
/*   genrand() generates one pseudorandom real number (double)  */
/* which is uniformly distributed on [0,1)-interval, for each   */
/* call. sgenrand(seed) sets initial values to the working area */
/* of 624 words. Before genrand(), sgenrand(seed) must be       */
/* called once. (seed is any 32-bit integer.)                   */
/* Integer generator is obtained by modifying two lines.        */
/*   Coded by Takuji Nishimura, considering the suggestions by  */
/* Topher Cooper and Marc Rieffel in July-Aug. 1997.            */

/* This library is free software; you can redistribute it and/or   */
/* modify it under the terms of the GNU Library General Public     */
/* License as published by the Free Software Foundation; either    */
/* version 2 of the License, or (at your option) any later         */
/* version.                                                        */
/* This library is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            */
/* See the GNU Library General Public License for more details.    */
/* You should have received a copy of the GNU Library General      */
/* Public License along with this library; if not, write to the    */
/* Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA   */ 
/* 02111-1307  USA                                                 */

/* Copyright (C) 1997, 1999 Makoto Matsumoto and Takuji Nishimura. */
/* Any feedback is very welcome. For any question, comments,       */
/* see http://www.math.keio.ac.jp/matumoto/emt.html or email       */
/* matumoto@math.keio.ac.jp                                        */

/* REFERENCE                                                       */
/* M. Matsumoto and T. Nishimura,                                  */
/* "Mersenne Twister: A 623-Dimensionally Equidistributed Uniform  */
/* Pseudo-Random Number Generator",                                */
/* ACM Transactions on Modeling and Computer Simulation,           */
/* Vol. 8, No. 1, January 1998, pp 3--30.                          */


/* Period parameters */  
#define N 624
#define M 397
#define MATRIX_A 0x9908b0df   /* constant vector a */
#define UPPER_MASK 0x80000000 /* most significant w-r bits */
#define LOWER_MASK 0x7fffffff /* least significant r bits */

/* Tempering parameters */   
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)

static unsigned long mt[N]; /* the array for the state vector  */
static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */

/* Initializing the array with a seed */

void sgenrand(seed)
    unsigned long seed;	
{
    int i;

    for (i=0;i<N;i++) {
         mt[i] = seed & 0xffff0000;
         seed = 69069 * seed + 1;
         mt[i] |= (seed & 0xffff0000) >> 16;
         seed = 69069 * seed + 1;
    }
    mti = N;
}

/* Initialization by "sgenrand()" is an example. Theoretically,      */
/* there are 2^19937-1 possible states as an intial state.           */
/* This function allows to choose any of 2^19937-1 ones.             */
/* Essential bits in "seed_array[]" is following 19937 bits:         */
/*  (seed_array[0]&UPPER_MASK), seed_array[1], ..., seed_array[N-1]. */
/* (seed_array[0]&LOWER_MASK) is discarded.                          */ 
/* Theoretically,                                                    */
/*  (seed_array[0]&UPPER_MASK), seed_array[1], ..., seed_array[N-1]  */
/* can take any values except all zeros.                             */

void lsgenrand(seed_array)
    unsigned long seed_array[];
    /* the length of seed_array[] must be at least N */
{
    int i;

    for (i=0;i<N;i++) 
      mt[i] = seed_array[i];
    mti=N;
}

double /* generating reals */
/* unsigned long */ /* for integer generation */

genrand()
{
    unsigned long y;
    static unsigned long mag01[2]={0x0, MATRIX_A};
    /* mag01[x] = x * MATRIX_A  for x=0,1 */

    if (mti >= N) { /* generate N words at one time */
        int kk;

        if (mti == N+1)   /* if sgenrand() has not been called, */
            sgenrand(4357); /* a default initial seed is used   */

        for (kk=0;kk<N-M;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        for (;kk<N-1;kk++) {
            y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
            mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
        }
        y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
        mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

        mti = 0;
    }
  
    y = mt[mti++];
    y ^= TEMPERING_SHIFT_U(y);
    y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
    y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
    y ^= TEMPERING_SHIFT_L(y);

    return ( (double)y * 2.3283064365386963e-10 ); /* reals: [0,1)-interval */
    /* return y; */ /* for integer generation */
}


/* This main() outputs first 1000 generated numbers.  */
//main()
//{ 
//    int i;
//
//    sgenrand(4357);
//    for (i=0; i<1000; i++) {
//        printf("%10.8f ", genrand());
//        if (i%5==4) printf("\n");
//    }
//}

// ============ End of Twister ===========



/* --- L'Ecuyer RN generators --- */ 

 #define norm 2.328306549295728e-10
 #define m1   4294967087.0
 #define m2   4294944443.0
 #define a12     1403580.0
 #define a13n     810728.0
 #define a21      527612.0
 #define a23n    1370589.0

 double  s10, s11, s12, s20, s21, s22;

 void InitMRGa (long seed)
 {
    double urand[6];
    long itemp;

/*---- itemp preset of starting values to test generator ---- */
/*
    urand[0] = 0.1;
    urand[1] = 0.3;
    urand[2] = 0.5;
    urand[3] = 0.6;
    urand[4] = 0.7;
    urand[5] = 0.9;


    // printf("seed = %ld\n",seed);
    // getch();

    itemp = m1 * urand[0]; s10 = itemp;
    itemp = m1 * urand[1]; s11 = itemp;
    itemp = m1 * urand[2]; s12 = itemp;

    itemp = m2 * urand[3]; s20 = itemp;
    itemp = m2 * urand[4]; s21 = itemp;
    itemp = m2 * urand[5]; s22 = itemp;
*/
    itemp = seed; s10 = itemp;
    itemp = 3; s11 = itemp;
    itemp = 5; s12 = itemp;

    itemp = 6; s20 = itemp;
    itemp = 7; s21 = itemp;
    itemp = 9; s22 = itemp;


 }

 double MRGa ()
/*WAS MRG32k3a*/
    {
    long   k;
    double p1, p2;
    /* Component 1 */
    p1 = a12 * s11 - a13n * s10;
    k = p1 / m1;   p1 -= k * m1;   if (p1 < 0.0) p1 += m1;
    s10 = s11;   s11 = s12;   s12 = p1;
    /* Component 2 */
    p2 = a21 * s22 - a23n * s20;
    k  = p2 / m2;  p2 -= k * m2;   if (p2 < 0.0) p2 += m2;
    s20 = s21;   s21 = s22;   s22 = p2;
    /* Combination */
    if (p1 <= p2) return ((p1 - p2 + m1) * norm);
    else return ((p1 - p2) * norm);
    }

long zrng = 1234567;

#define MULT1 24112
#define MULT2 26143
#define MODLUS 2147483647

double urand()
{
long zi,lowprd,hi31;
double x;

zi = zrng;
lowprd =(zi & 65535) *MULT1;
hi31 = (zi>> 16)*MULT1 + (lowprd >> 16);
zi = ((lowprd &65535) - MODLUS) +
     ((hi31 & 32767) << 16) + (hi31 >> 15);
if (zi < 0) zi += MODLUS;
lowprd =(zi & 65535) *MULT2;
hi31 = (zi>> 16)*MULT2 + (lowprd >> 16);
zi = ((lowprd &65535) - MODLUS) +
     ((hi31 & 32767) << 16) + (hi31 >> 15);
if (zi < 0) zi += MODLUS;
zrng = zi;


x = ((zi>>7 | 1) + 1)/16777216.0;

/*
x = zi/(double)MODLUS;
*/

return x;
}



int main(int argc, char * argv[])

{/*0*/

 extern double neglikval(double *, double *);
 extern double neglogposterior(double *, double *);

 extern void sortintoascend3(long, double*, double*, double* );
 extern void sortintoascend2(long, double*, double* );
 extern void sortintoascend1(long, double*);

 extern void sortintodescend2(long, long, double*, double* );

 extern void ShellSortintoAscendingOneCol(long, double* );

 extern void ShellSortintoAscendingTwoCols(long, long, double*, double* );

 extern void ShellSortintoAscendingThreeCols(long, double*, double*, double* );

 extern void ShellSortintoDescendingTwoCols(long, long, double*, double* );


 extern double xinvnm(double);


 extern long CatSample( double *, long);

 extern void HistoOnly(
             double xorig[][BSIZE],  /* IN: the points in columns */
             double yorig[],         /* IN: the postprob in a column */
             long numintvls,              /* IN: # of histogram cells */
             double freq[][MAXNUMINTVLS], /* OUT: cell frequencies */
             long numatk[],               /* OUT: number of obsns */
             double hmax[],               /* OUT: max histo height */
             long numpars,                /* IN: # of parameters being done */
             long numpoints,              /* IN: # of points */

//             double fval[][MAXNUMINTVLS], /* NOT USED: fitted pdf */
//             double parc[][PSIZE],    /* NOT USED: parameters of pdfs */
             double ProbofK[], 
             double SDProbofK[], 
             FILE * fopdf /* IN/OUT */
          );

#if GRAPHICSON

 extern void HistoOnlyB(
             double xorig[][BSIZE],  /* IN: the points in columns */
             double yorig[],         /* IN: the postprob in a column */
             long numintvls,              /* IN: # of histogram cells */
             double freq[][MAXNUMINTVLS], /* OUT: cell frequencies */
             long numatk[],               /* OUT: number of obsns */
             double hmax[],               /* OUT: max histo height */
             long numpars,                /* IN: # of parameters being done */
             long numpoints,              /* IN: # of points */

//             double fval[][MAXNUMINTVLS], /* NOT USED: fitted pdf */
//             double parc[][PSIZE],    /* NOT USED: parameters of pdfs */
             long k,          /* ####### additional input parameter  */
             double ProbofK[],
             long firsttimeflag /* IN: Set to 1 if HistoOnly called only once */
          );

#endif

 extern  double logCandPDF1 (
        long NumCandiTries[],
        long iattempt,
        long model,
	double FittedMeanAll[][TWO][MAXNUMCOMPS],
	double FittedTauAll[][TWO][MAXNUMCOMPS],
	long k,
        long jtry,
	double FittedDeltaAll[][TWO][MAXNUMCOMPS],

        /* double sampleBeta, */

        double nudof,
	double sampleLsAll[][PSIZE][TWO][MAXNUMCOMPS],
        long bestjtryAll [],

        /* double tauConst, */

        double mu[],
	double tau[],
	double awt[],
        long i,
        long outnum, // no longer used
        double stDevT,
        double invsamLsAll[][PSIZE][TWO][MAXNUMCOMPS],
        double mu1[],
        double tau1[],
        double awt1[]
	);

 extern double Determinant(long dimension, double matrix[][PSIZE]);

 extern double gaussj(long nn, 
           double a[][PSIZE], double inverse_a[][PSIZE]);

 FILE * fp9;

static double logFactorial[MAXNUMCOMPS];
static long NumCandiTries[MAXNUMCOMPS];
static long NumCandiReplis[MAXNUMCOMPS];

static double logTriesRatio[MAXNUMCOMPS];

static double param[PSIZE];
static double probK[MAXNUMCOMPS];

double priorAlpha;
double priorBeta;
double priorDelta;
double priorKappa;
double priorNu1;
double priorNu2;
double priorGamma2;
long priorKMax;
double priorXi;
double Groupsize;
double Stretch;
long seedIS;

static double invsamLsAll[PSIZE][PSIZE][TWO][MAXNUMCOMPS];
static double CandWt[TWO][MAXNUMCOMPS];
static double cov[PSIZE][PSIZE];
static double invcov[PSIZE][PSIZE];
static double CondLogCandProb[TWO];

long iattempt;

long show;

double nudof, stDevT;

double determ;
double LogCandProb;
double LogPostProb;
double Meanlogratio;
double MaxLogRatio;
double MinLogRatio;
long NumPosRatios;

static double mu [ MAXNUMCOMPS ];
static double tau[ MAXNUMCOMPS ];
static double awt[ MAXNUMCOMPS ];

static double mu1 [ MAXNUMCOMPS ];
static double tau1[ MAXNUMCOMPS ];
static double awt1[ MAXNUMCOMPS ];

static double kArray   [ BSIZE];

static double logRatio [ BSIZE];
static double Ratio    [ BSIZE];
static double postProb [ BSIZE];

static double ProbofK[MAXNUMCOMPS];
static double SDProbofK[ MAXNUMCOMPS];

static double ProbofK2[ MAXNUMCOMPS];
static double SDProbofK2[ MAXNUMCOMPS];


static long PosProbFlag[ BSIZE];


double Psum;

static double AACandProb   [BSIZE];
static double AAPostProb   [BSIZE];
static double AACompNum    [BSIZE];
static double AALogCandProb[BSIZE];
static double AALogPostProb[BSIZE];
static double AALogRatio   [BSIZE];

static double C4Cells[BSIZE][PSIZE];
static double TempCells[BSIZE][PSIZE];
static double Asort[BSIZE];
static double Bsort[BSIZE];
static NumWithKcomps[PSIZE];

FILE * fp8;
long out8flag;

 long numcomps0;
 long outnum;
 long numoISsamples;

 // double sampleBeta;

 double minDelta;

 // double tauConst; // not used in this version of IS.

  static double xorig[PSIZE][BSIZE];  /* IN: the points in columns */
  static double yorig[BSIZE];  /* IN: the postprobs in a column */
  long numintvls;         /* IN: # of histogram cells */
  static double freq[PSIZE ][MAXNUMINTVLS]; /* OUT: cell frequencies */
  static long numatk[PSIZE];          /* OUT: number of obsns */
  static double hmax[MAXNUMINTVLS];   /* OUT: max histo height */
  long numpars;            /* IN: # of parameters being done */
  long numpoints;     /* IN: # of points */
  long numpoints0;    /* IN: # of points for reduced plot */


  long firsttimeflag;      /* IN: Set to 1 if HistoOnly called only once */

// void addinterval(long, long, double*, double*, double*, long*, long*, long*, long*);

// double explicitcdf( double, double*);

 double gammalog(double, long);
 double bigphi(double);
 double incompgam(double, double, double, long);

 long adjustonflag;
 double adstat;

 FILE * fp1; // input file
 FILE * fout;
 FILE * fout0;
 FILE * fp5;

 FILE * fopdf;


static long akey;

static long ib, iis, m, p;
static long jk;
static long istt, iend;
static long istt0, iend0;

static double xd, xt;
static double yt;

static double aicstat;
static double bicstat;

static long bestjtry;
static double dummy;
static long lessjtry;
static double lessratio;

static long kMaxSample;


 static long i;
 static long inel;
 static long ifault=0;
 static long igiven;

 static long j;
 static long j0, j1, j2, j3;
 static long jlo, jhi;
 static long j0old,j1old;
 static long jtry;

 static long k;
 
 static double lam;

 static double maxvalue, maxvaluesofar;
 static double maxprobsofar, maxprobsofar0, maxprobsofar1;
 static double mux,sig;

 static long narrowflag;
 static double newlikelihdval, newspacingsval;
 static long  numcompsprev, numcompspres, numcompsnew;
 static long numtries;
 static long nxplct;

 static long numGraphs;
 static long numpos[51];
 static long numpos1[51];

 static double obsmean, obssig;
 
 static long onepass;

 static long pauseflag;
 static double probk;

 static double sampmean,sampsig,sampskew,samprange;
 static double samplogmean, samplogsig, samphmean;
 
 

 static double WeibullAlpha, WeibullCee;
 
 static double x0, x1, x2;


 static double tee;
 static double thetamas;

 static double upper;

 static double yobsn[NSIZE];
 static double yobsno[NSIZE];

 static double priorpar[PSIZE];

static long it, jt, iquotient, jquotient, iremainder, jremainder;

double ShrinkFactor;

static double sampleL[PSIZE][PSIZE];
static double sampleLsAll[PSIZE][PSIZE][TWO][MAXNUMCOMPS];

static double ISmu[MAXNUMCOMPS][MAXNUMCOMPS];
static double IStau[MAXNUMCOMPS][MAXNUMCOMPS];
static double ISawt[MAXNUMCOMPS][MAXNUMCOMPS];

static double FittedMean[MAXNUMCOMPS][MAXNUMCOMPS];
static double FittedTau[MAXNUMCOMPS][MAXNUMCOMPS];
static double FittedAwt[MAXNUMCOMPS][MAXNUMCOMPS];

static double FittedMeanAll[MAXNUMCOMPS][TWO][MAXNUMCOMPS];
static double FittedTauAll[MAXNUMCOMPS][TWO][MAXNUMCOMPS];
static double FittedAwtAll[MAXNUMCOMPS][TWO][MAXNUMCOMPS];

static double AICStatAll [TWO][MAXNUMCOMPS];
static double BICStatAll [TWO][MAXNUMCOMPS];

static long bestjtryAll      [MAXNUMCOMPS];
static double maxvalsAll     [TWO][MAXNUMCOMPS];
static double infvarsAll [PSIZE][PSIZE][TWO][MAXNUMCOMPS];

double x;
double maxlogpostprob;

static char DataFileName[20];
static char DataFolderInit[30];
static char DataFolderOut7[30];
static char DataFolderOut8[30];
static char DataFolderOut9[30];
static char DataFolderOutPdf[30];


//  ======== read input data from Init.d and Out9.d  ===============

yobs = yobsn;
show = 0;

strcpy(DataFileName, argv[1]);

strcpy(DataFolderInit,DataFileName);
strcat(DataFolderInit,"/Init.d");

strcpy(DataFolderOut7,DataFileName);
strcat(DataFolderOut7,"/Out7.d");

strcpy(DataFolderOut8,DataFileName);
strcat(DataFolderOut8,"/Out8.d");

strcpy(DataFolderOut9,DataFileName);
strcat(DataFolderOut9,"/Out9.d");

strcpy(DataFolderOutPdf,DataFileName);
strcat(DataFolderOutPdf,"/outpdf.d");

printf("datafile name = %s\n",DataFileName);
printf("Input file = %s\n",DataFolderInit);
printf("Input file = %s\n",DataFolderOut9);
printf("Output file = %s\n",DataFolderOut7);
printf("Output file = %s\n",DataFolderOut8);

#if OUTPUTPDF
   printf("Output file = %s\n",DataFolderOutPdf);
#endif

printf("\n");

//printf("\nPress a key to continue");
//getch();


fp1 = fopen(DataFolderInit,"r");
fscanf(fp1,"%ld",&n);
for (i=1;i<=n;i++)
{/*2*/
   fscanf(fp1,"%lf\n",&yobsn[i]); // --- redundant read
}/*2*/

fscanf(fp1,"%lf", &x); // --- dummy read: Dataspread not used in imp.c
fscanf(fp1,"%ld", &i); // --- dummy read: model not used in imp.c
// fscanf(fp1,"%ld",&i); // --- dummy read:  outnum not used in imp.c
fscanf(fp1,"%ld", &numoISsamples); // --- This is one purpose of reading Init.d
fscanf(fp1,"%lf", &x); // mindelta
// fscanf(fp1,"%lf", &x); // tauconst // --- removed from mixture51
fscanf(fp1,"%lf", &x); //priordelta
fscanf(fp1,"%ld", &i); //priorkmax
fscanf(fp1,"%lf", &x); //priornu1

fscanf(fp1,"%lf", &x); //stretch
fscanf(fp1,"%ld", &x); //groupsize

fscanf(fp1,"%lf", &x); //prioralpha
fscanf(fp1,"%lf", &x); //priorgee
fscanf(fp1,"%lf", &x); //priorhch


fscanf(fp1,"%ld", &seedIS); //seedIS --- This is the other purpose of reading Init.d


fclose(fp1);

fp9 = fopen(DataFolderOut9,"r");
fscanf(fp9,"%ld", &n); // ---- n is the sample size ----
for(i=1;i<=n;i++)
{
   fscanf(fp9,"%lf", &yobsn[i]);
//   printf("y[%ld]=%lf\n",i, yobsn[i]);
//   getch();
}

// sortintoascend1(n, yobsn);

ShellSortintoAscendingOneCol(n, yobsn );

//for(i=1;i<=n;i++)
//{
//   printf("y[%ld]=%lf\n",i, yobsn[i]);
//   getch();
//}


fscanf(fp9,"%ld", &model);
fscanf(fp9,"%ld", &i); // Dummy. This was numoISsamples - now read from Init.d

// fscanf(fp9,"%lf", &sampleBeta);
// printf("sampleBeta = %e\n",sampleBeta);

fscanf(fp9,"%lf", &minDelta );
// printf("minDelta = %e\n",minDelta);

// fscanf(fp9,"%lf", &tauConst);

// printf("tauConst = %e\n",tauConst);

for(i=1; i<=17; i++)
{
   fscanf(fp9,"%lf", &priorpar[i]);
//   printf("priorpar[%ld]=%e\n",i,priorpar[i]);
}

printf("\n\n");

printf("==========================================================\n\n");

priorKMax = priorpar[6];
switch(model)
 {
    case XMIXTURE: printf("XMIXTURE (Extreme value mixture))\n"); break;
    case NXMIXTURE: printf("NXMIXTURE (Negative extreme value mixture)\n"); break;
    case WMIXTURE: printf("WMIXTURE (Weibull mixture)\n"); break;
    case NMIXTURE: printf("NMIXTURE (Normal mixture)\n"); break;
    case LMIXTURE: printf("LMIXTURE (Lognormal mixture)\n"); break;
    case GMIXTURE: printf("GMIXTURE (Gamma mixture)\n"); break;
    case IMIXTURE: printf("IMIXTURE (Inverse Gaussian mixture)\n"); break;
 }
printf("\nmodel ID = %ld\n",model);

printf("priorKMax = %ld\n",priorKMax);

numcomps0 = priorpar[6];
// printf("numcomps0 = %ld\n",numcomps0); getch();


for(k=1; k<=numcomps0; k++)
{
   for(jtry=0; jtry<TWO; jtry++)
   {
      for(j=1; j<=k; j++)
      {
         fscanf(fp9,"%lf", &FittedMeanAll[j] [jtry] [k] );
         fscanf(fp9,"%lf", &FittedTauAll [j] [jtry] [k] );
         fscanf(fp9,"%lf", &FittedAwtAll [j] [jtry] [k] );
      }

      // --- these are just passed through to out8.d
      fscanf(fp9,"%lf", &AICStatAll[jtry] [k]); // This is the read of the aicstat
      fscanf(fp9,"%lf", &BICStatAll[jtry] [k]); // This is a the read of the bicstat

   }
}

logFactorial[0] = 0.0;
for(k=1; k<=numcomps0; k++)
{
   fscanf(fp9,"%ld", &bestjtryAll [k] );
//   printf("bestjtryAll[%ld]=%ld\n",k, bestjtryAll[k]);
//   getch();
   NumCandiTries[k] = 0;
   NumCandiReplis[k] = 0;
   logFactorial[k] = logFactorial[k-1] + log( (double)k );
}

#if 1
for(k=1; k<=numcomps0; k++)
{
   for(jtry=0; jtry<TWO; jtry++)
   {
      fscanf(fp9,"%lf", &maxvalsAll [jtry] [k] );
//      printf("jtry=%ld k=%ld max=%lf\n",jtry,k,maxvalsAll[jtry][k]);
   }
}
#endif

for(k=1; k<=numcomps0; k++)
{
   numpars = 3*k;
   for(i=1;i<=numpars;i++)
   {
      for(j=1;j<=numpars;j++)
      {
         for(jtry=0; jtry<TWO; jtry++)
         {
            fscanf(fp9,"%lf\n", &infvarsAll [i][j][jtry][k] ); 
            fscanf(fp9,"%lf\n", &sampleLsAll[i][j][jtry][k] ); 
         }
      }
   }
}

fclose(fp9);

//printf("At a seed= %ld\n",seedIS);
//getch();


// ----- init for LEcuyer mRG RN generator
// InitMRGa(seedIS);

//printf("At a s10= %lf\n",s10);
//printf("At a s11= %lf\n",s11);
//printf("At a s12= %lf\n",s12);
//printf("At a s20= %lf\n",s20);
//printf("At a s21= %lf\n",s21);
//printf("At a s22= %lf\n",s22);
//getch();

// ------ Init for Twister RN generator --------

if(seedIS==0)
{

seedagain: 
  // seedIS = (unsigned long) time(0);
  seedIS = (unsigned long) 12345;

  if(seedIS==0) goto seedagain;

  printf("\nrandomized seed = %ld\n\n", seedIS);

}
else
{
  printf("\nselected seed = %ld\n\n", seedIS);

}


sgenrand(seedIS);

// ------ diagnostic -----
//    for (i=0; i<1000; i++) {
//        printf("%10.8f ", genrand());
//        if (i%5==4) printf("\n");
//    }
//  getch();

//priorAlpha = priorpar[1];
//priorBeta  = priorpar[2];
//priorKappa = priorpar[3];
//priorXi    = priorpar[4];
//priorDelta = priorpar[5];

priorKMax  = priorpar[6];

//Groupsize  = priorpar[11];

// Adjustment to ensure FittedAwt's are at least .01 and sum to one
/*==============
for(k=1; k<=priorKMax; k++)
{
    for(jtry=0; jtry<TWO; jtry++)
    for(j=1; j<=k; j++)
    {
        if(FittedAwtAll[j][jtry][k] < minDelta)
        {
            FittedAwtAll[j][jtry][k] = FittedAwtAll[j][jtry][k] + minDelta;
            for(i=1; i<j; i++)
            {
                FittedAwtAll[i][jtry][k] = FittedAwtAll[i][jtry][k] - minDelta / (k - 1.);
            }
            for(i=j+1; i<=k; i++)
            {
                FittedAwtAll[i][jtry][k] = FittedAwtAll[i][jtry][k] - minDelta / (k - 1.);
            }
        }
    }
}
===============*/

for(i=1; i<=priorKMax; i++)
{
   probK[i] = 1. / priorKMax;
}

NumPosRatios = 0.;
Meanlogratio = 0.;
MaxLogRatio = -1E+20;
MinLogRatio = 1E+20;

printf("Number of IS samples = %ld\n\n",numoISsamples);

printf("\nPress a key to start Importance Sampling, Output is in Out7.d & Out8.d\n");
getch();

printf("Importance Sampling Started\n");


out8flag=1; // outflag = 0 to suppress output

// calc invsamLsALL

#if 1
for(k=1; k<=priorKMax; k++)
{
   // --- calculate Candidate weights
   // printf("At k = %ld\n",k);

   bestjtry = bestjtryAll[k];
   lessjtry = 1-bestjtry;

   lessratio = maxvalsAll[lessjtry][k] - maxvalsAll[ bestjtry ] [k];

   if( lessratio< -300.0) lessratio = -300.0;
   if( lessratio> 300.0) lessratio = 300.0;

   //printf("At k = %d: bestjtry=%ld lessjtry=%ld lessratio = %le\n",
   //          k, bestjtry,lessjtry,lessratio);
   //getch();

   lessratio = exp(lessratio);


   // do not bother with smaller term if it is less than .001 of best
   if(lessratio<0.001) lessratio = 0.0;
   CandWt[bestjtry][k] = 1.0/(1.0+lessratio);
   CandWt[lessjtry][k] = lessratio/(1.0+lessratio);
   // printf("CandWtbest = %lf at jtry = %ld\nCandWtless = %lf at jtry = %ld\n",
   //        CandWt[bestjtry][k], bestjtry, CandWt[lessjtry][k], lessjtry);
   // getch();

   // calculate inverse of 'lessjtry' term provided it exists
   jtry = lessjtry;
   // --- calc inverse of L's
   for(i=1; i<=3*k-1; i++) 
   {
      {
         for(j=1; j<=3*k-1; j++) 
         cov[i][j] = sampleLsAll[i][j][jtry][k];
      }
   }
      
    determ = Determinant(3*k-1, cov);
    // getch();

   if(fabs(determ) > 0.0 ) // ok to calculate inverse
   {
      gaussj(3*k-1, cov, invcov);

      /* --------------
      for(i=1; i<=3*k-1; i++) 
      {
         for(j=1; j<=3*k-1; j++)
         {
            printf("ic[%2d][%2d]=%6f ",i,j,invcov[i][j]);
            if(j%6==0) printf("\n");
         }
         printf("\n");
      }
      getch();
      -------------- */
      for(i=1; i<=3*k-1; i++) 
      {
         for(j=1; j<=3*k-1; j++) 
         invsamLsAll[i][j][jtry][k] = invcov[i][j];
      }
   }
   else // covariance not satisfactory so do not attempt second term
   {
      CandWt[bestjtry][k] = 1.0;
      CandWt[lessjtry][k] = 0.0;
   }
}
#endif

// -------- Calculate candidate prob parameters ---------
nudof = 3.0;
stDevT = sqrt( nudof/( nudof-2.) ); // this is the sd of the tdistn
numevals = 0;

iis = numoISsamples/numcomps0 + 1; // find whole number of repeats of full cycle submodels to be sampled 
numoISsamples = iis * numcomps0;

k = 0;
// ================ MAIN Importance Sampling Loop ========================
for(iis=1; iis<=numoISsamples; iis++)
{/*iis*/

    if (iis % 500 == 0 ) printf("numevals = %d\n", iis);

    // set value of k (number of components in model), cyclically
    //k = CatSample(probK, priorKMax);

    k = k+1;
    if (k > numcomps0) k = 1;

    numcomps = k; // numcomps is a global variable, this allows it to
                  // be accessed by other external routines

    // printf("numcomps = %ld\n",numcomps);
    // getch();


    // --- select jtry
    xd = MRGa();
    bestjtry = bestjtryAll[k];

    CandWt[bestjtry][k] = 1.0; // WATCHPOINT this is set to unity
                 // so that only besttry is ever selected in what follows

    if(xd < CandWt[ bestjtry ][k]) jtry = bestjtry;
    else jtry = 1- bestjtry;

    // printf("k=%ld bestjtry=%ld xd=%lf CandWt=%lf \n",
    //               k, bestjtry,xd, CandWt[bestjtry][k]);

    // --- calc conditional density for jtry
    iattempt = 1;

        xt = CandWt[jtry][k];
        if(xt==0.0) {printf("At 0: xt = %lf\n"); getch();}

    CondLogCandProb[0] =  log(CandWt[jtry][k]) +
                      logCandPDF1(NumCandiTries, iattempt, model,
                FittedMeanAll, FittedTauAll,
                k, jtry, FittedAwtAll, /* sampleBeta, */ nudof, sampleLsAll,
                bestjtryAll, /* tauConst,*/ mu, tau, awt, iis, outnum,
                stDevT, invsamLsAll, mu1, tau1, awt1 );

    // Note invsamLsall, mu1, tau1, awt1 are all inputs
    // but are not used for this iattempt.
    // However in this first attempt mu, tau and awt are outputs
    // printf("\nCondlogcandpdf0=%lf\n", CondLogCandProb[0]);

    for(j=1; j<=k; j++)
    {
	if (mu[j]<=0.000001) mu[j] = .000001;
	if (tau[j]<=0.000001) tau[j] = .000001;
	if (awt[j]<=0.000001) awt[j] = .000001;

        param[3*(j-1)+1] = mu[j];
        param[3*(j-1)+2] = tau[j]; // param is sigma and tau is also sig
        param[3*(j-1)+3] = awt[j];

        /*==========
        {
        printf("iis = %ld mu[%ld] = %lf tau = %lf wt = %lf\n",iis, j, mu[j], tau[j], awt[j]);
        getch();

        }
        ===========*/
    }

    LogPostProb = -neglogposterior( param, priorpar) + logFactorial[k];


    LogCandProb = CondLogCandProb[0];


    // {printf("At 2 iis=%ld : logpostprob = %lf logcandprob = %lf\n",
    //                    iis, LogPostProb, LogCandProb); getch();}

    // ---- Note: above CandProb does not yet include the acceptance/rejection normalization


//    if( LogPostProb > -10000 ) // ----- This non zero prob WAS -10000

    if(LogPostProb - LogCandProb > -100000)
    {
        logRatio[iis] = LogPostProb - LogCandProb; // not A/R normalized

        // printf("at if2: logratio = %lf\n", logRatio[iis]);
        // getch();

        NumPosRatios = NumPosRatios + 1;
        PosProbFlag[iis] = 1;
    }
    else // --- This is a zero probability case
    {
        // logRatio[iis] = LogPostProb;

        logRatio[iis] = -10000;

        // printf("at else2: logratio = %lf\n", logRatio[iis]);
        // getch();

        PosProbFlag[iis] = 0;
    }

   if(out8flag==1)
   {
      // --- This is not needed as the positions of the components are fixed
      //     by the optimization process.

      // ---Sort using means

      // sortintoascend3(k, mu, tau, awt);

      // ShellSortintoAscendingThreeCols (k, mu, tau, awt);
    
      // store IS output    
      AACandProb[iis] = exp(LogCandProb);
      // AAPostProb[iis] = exp(LogPostProb); // set later after offset subtracted
      AACompNum[iis] = k;
      AALogCandProb[iis] = LogCandProb;
      AALogPostProb[iis] = LogPostProb;
      // AALogRatio[iis] = LogPostProb - LogCandProb; //set later

      // --- put parameters and ratio in the worksheet lookalike
      C4Cells[iis] [1] = iis;
      C4Cells[iis] [3] = k;
      for(j=1; j<=k; j++)
      {
         C4Cells[iis][j + 3] = mu[j];
         C4Cells[iis][j + 3 + priorKMax] = tau[j];
         C4Cells[iis][j + 3 + 2 * priorKMax] = awt[j];
         // printf("mu=%lf tau=%lf awt=%lf\n",mu[j],tau[j],awt[j]);
      }
      // getch();

    }/*out8flag==1*/
    
}// Next iis


// ===== Find number of samples with k components, for each k =====
for(j=1; j<=priorKMax; j++)
{
  NumWithKcomps[j] = 0;
}
for(iis=1; iis<=numoISsamples; iis++)
{
   if( PosProbFlag[iis] == 1)
   {
       k = AACompNum[iis];
       NumWithKcomps[k]++;
   }
}
for(j=1; j<=priorKMax; j++)
{
      printf("Numsamples with %ld comps = %ld\n",j, NumWithKcomps[j]);
}

// ======= calculate Acceptance/Rejection scaling factor for the Candi probs of each k =====
for(k=1; k<=priorKMax; k++)
{ // ----- This is log(# points at k / # tries at k)
    logTriesRatio[k] = log ( (double)NumCandiTries[k] / (double)NumWithKcomps[k]);
    // printf("logtriesratio[%ld]=%lf\n", k, logTriesRatio[k]);
}
// getch();
// =================================

maxlogpostprob = -1e20; 
for(iis=1;iis<=numoISsamples;iis++)
{
   if(AALogPostProb[iis]>maxlogpostprob) maxlogpostprob = AALogPostProb[iis];

   // include simplex normalization of the candiprob and logcandiprob
   // (makes them larger)
   k = AACompNum[iis];

   AACandProb[iis] = exp( logTriesRatio[k] ) * AACandProb[iis];

   AALogCandProb[iis] = AALogCandProb[iis] + logTriesRatio[ k ];

   // include simplex normalization of logRatio (makes it smaller)
   logRatio[iis]  = logRatio[iis] - logTriesRatio[k];
   
   if(PosProbFlag[iis]==1) Meanlogratio = Meanlogratio + logRatio[iis];
}


for(iis=1;iis<=numoISsamples;iis++)
{
      AALogPostProb[iis] = AALogPostProb[iis]-maxlogpostprob; // LogPostProb - const

      AAPostProb[iis] = exp( AALogPostProb[iis] ); // PostProb/MaxPostProb

      AALogRatio[iis] = AALogPostProb[iis] - AALogCandProb[iis]; // LogRatio

      if (logRatio[iis] > MaxLogRatio)
      {
         MaxLogRatio = logRatio[iis];
      }
//      if (logRatio[iis] < MinLogRatio)
//      {
//         MinLogRatio = logRatio[iis];
//      }


}

printf("Maxlogratio = %le\n",MaxLogRatio);
// printf("Minlogratio = %le\n",MinLogRatio);

// printf("Meanlogratio = %le\n",Meanlogratio);

if( NumPosRatios > 0)
{
   Meanlogratio = Meanlogratio / NumPosRatios;
   printf("Meanlogratio = %lf\n",Meanlogratio);
}

printf("\n Processing results, please wait\n\n");

/* ====== Do the Bayes normalization and calc posterior prob of each sample ===== */

// ----- reset logratios

for( i = 1; i<= numoISsamples; i++)
{
    if( PosProbFlag[i] == 1)
    {
        Ratio[i] = exp(logRatio[i] - MaxLogRatio);
    }
    else
    {
        Ratio[i] = 0.0;
    }

}

Psum = 0.0;
// ----- Sum the ratios of all the IS observations
for(i=1; i<=numoISsamples; i++)
{
   Psum = Psum + Ratio[i];
}

//printf("Psum=%le\n",Psum);
//getch();


// ----- Scale so that they are probabilities
for(i=1; i<=numoISsamples; i++)
{
        postProb[i] = Ratio[i] / Psum;
	if(out8flag==1) C4Cells[i][ 2] = postProb[i];
}        

/* ----- Sort data samples according to number of components in sample ----- */
for(i=1; i<=numoISsamples; i++)
{
        Asort[i] = C4Cells[i][3]; // these are the k's (number of components)
        Bsort[i] = i;
}

// sortintoascend2(numoISsamples, Asort, Bsort);

ShellSortintoAscendingTwoCols(1, numoISsamples, Asort, Bsort);


/* ----- Save the sorted IS results -------- */
for(i=1; i<=numoISsamples; i++)
{
        k = Bsort[i];
        TempCells[i][1] = AACandProb   [k];
        TempCells[i][2] = AAPostProb   [k];
        TempCells[i][3] = AACompNum    [k];
        TempCells[i][4] = AALogCandProb[k];
        TempCells[i][5] = AALogPostProb[k];
        TempCells[i][6] = AALogRatio   [k];
}

for(i=1; i<=numoISsamples; i++)
{
        AACandProb   [i] = TempCells[i][1];
        AAPostProb   [i] = TempCells[i][2];
        AACompNum    [i] = TempCells[i][3];
        AALogCandProb[i] = TempCells[i][4];
        AALogPostProb[i] = TempCells[i][5];
        AALogRatio   [i] = TempCells[i][6];
}

for(i=1; i<=numoISsamples; i++)
{
        k = Bsort[i];
        for(j=2; j<=3+3*priorKMax; j++)
        {
           TempCells[i][j] = C4Cells[k][j];
        }
}
for(i=1; i<=numoISsamples; i++)
{
        for(j=2; j<=3+3*priorKMax; j++)
        {
           C4Cells[i][j] = TempCells[i][j];
        }
        kArray[i] = C4Cells [i] [3];
        postProb[i] = C4Cells [i] [2];
}
/* -------------------------------------------- */

/* ----------- now sort obsns for each number of components ---------- */
xt = 0.0;
istt = 1;
iend = NumWithKcomps[1];
for(jk=1; jk<=priorKMax; jk++)
{
   /* ------ For each k sort data by descending size of the postprob  ----- */
   for(i=istt; i<=iend; i++)
   {
      Asort[i] = -C4Cells[i][2]; // Postprob
      Bsort[i] = i;
   }

   //sortintodescend2(istt,iend,Asort,Bsort);

   ShellSortintoDescendingTwoCols (istt, iend, Asort, Bsort); // istt=1 is assumed

   //ShellSortintoAscendingTwoCols (istt, iend, Asort, Bsort);

   
   for(i=1; i<=iend; i++)
   {
      Asort[i] = -Asort[i]; // Postprob

//      printf("A[%ld]=%lf\n",i,Asort[i]);

   }
   


   /* ------ For each k save this sorted data  ----- */
   for(i=istt; i<=iend; i++)
   {

        k = Bsort[i];
        TempCells[i][1] = AACandProb   [k];
        TempCells[i][2] = AAPostProb   [k];
        TempCells[i][3] = AACompNum    [k];
        TempCells[i][4] = AALogCandProb[k];
        TempCells[i][5] = AALogPostProb[k];
        TempCells[i][6] = AALogRatio   [k];

   }
   for(i=istt; i<=iend; i++)
   {
        AACandProb   [i] = TempCells[i][1];
        AAPostProb   [i] = TempCells[i][2];
        AACompNum    [i] = TempCells[i][3];
        AALogCandProb[i] = TempCells[i][4];
        AALogPostProb[i] = TempCells[i][5];
        AALogRatio   [i] = TempCells[i][6];
   }

   for(i=istt; i<=iend; i++)
   {
      k = Bsort[i];
      for(j=2; j<=3+3*priorKMax; j++)
      {
         TempCells[i][j] = C4Cells[k][j];
      }
   }

   for(i=istt; i<=iend; i++)
   {
      for(j=2; j<=3+3*priorKMax; j++)
      {
         C4Cells[i][j] = TempCells[i][j];
      }
      kArray[i] = C4Cells [i] [3];
      postProb[i] = C4Cells [i] [2];
   }

   // ------ delete samples with excessively high postprob compared with next
   k = 0;
   xd = 0.0;
   for(i=istt+1; i<=iend; i++)
   {
      if( postProb[i-1] > REMOVEFACTOR * postProb[i] )
      {
         xd += postProb[i-1];
         postProb[i-1] = 0.0;
         C4Cells [i] [2] = 0.0;
         k++;
      }
      else
      {
         break;
      }
   }

   // ===== move array limits to next component ========

   istt += NumWithKcomps[jk];
   iend = istt + NumWithKcomps[jk+1] - 1;

   printf("At %ld components, %ld postprobs set to zero Value=%lf\n",
                                  jk, k, xd);

   xt += xd;

   // getch();

}/* jk */

// -------- Renormalise the postprobs that have not been deleted ---------
xd = 1.0 - xt;
//printf("xd= %le\n",xd);
//getch();

for(i=1;i<=numoISsamples;i++)
{
	postProb[i] = postProb[i]/xd;
        C4Cells[i][2] = postProb[i];
//printf("postprob[%ld]= %le\n",i,postProb[i]);
//getch();

}

// ============= MakeKFrequencyData ===============
for(i=1; i<=numoISsamples; i++)
{
    for(j=1; j<=priorKMax; j++)
    {
        if(kArray[i]==j)
        {
            ProbofK[j] = ProbofK[j] + postProb[i];
        }

    }
}

for(j=1; j<=priorKMax; j++)
{
    ProbofK2[j] = ProbofK[j]/NumWithKcomps[j];
}



// ============= Calc SDs for each ProbofK ===============

xt = 0.0;
for(j=1; j<=priorKMax; j++)
{
    SDProbofK[j] = 0.0;
    SDProbofK2[j] = 0.0;
    xt = xt + ProbofK[j];
}

//printf("Sum of probofK's = %lf\n", xt);
//getch();


for(i=1; i<=numoISsamples; i++)
{
    for(j=1; j<=priorKMax; j++)
    {
        if(kArray[i]==j)
        {

            SDProbofK[j] = SDProbofK[j] +
                  (1.-ProbofK[j]) * (1.-ProbofK[j]) * postProb[i] * postProb[i];

            SDProbofK2[j]  = SDProbofK2[j] + (postProb[i] - ProbofK2[j]) *
                                             (postProb[i] - ProbofK2[j]);
        }
        else
        {
            SDProbofK[j] = SDProbofK[j] +
                  ProbofK[j] * ProbofK[j] * postProb[i] * postProb[i];
        }

    }
}

for(j=1; j<=priorKMax; j++)
{
    SDProbofK[j] = pow( SDProbofK[j], 0.5);
    SDProbofK2[j] = pow( SDProbofK2[j], 0.5);

}


printf("\nProb Distn of # of Components:\n");
for(i=1; i<=priorKMax; i++)
{
    printf("prob[%ld]= %le SD= %le #Obsns= %ld #Tries= %ld\n",
          i, ProbofK[i], SDProbofK[i],
          NumWithKcomps[i], NumCandiTries[i]);
}

//printf("\n");

//for(i=1; i<=priorKMax; i++)
//{
//    printf(" prob[%ld] = %lf SD[%ld] = %lf\n",
//          i, ProbofK[i], i, SDProbofK[i]);
//}


if(out8flag==1)
// -------- Output all the sorted IS data to the Out8.d text file ---------
{
   fp8 = fopen(DataFolderOut8,"w");
   fprintf(fp8, "%ld\n", numoISsamples);
   fprintf(fp8, "%ld\n", priorKMax);
   fprintf(fp8, "%ld\n", model);

   for(j=1; j<=priorKMax; j++)
   {
      for(i=1; i<=j; i++)
      {
         bestjtry = bestjtryAll [j];
         fprintf(fp8, "%e ", FittedMeanAll[i] [bestjtry] [j] );
         fprintf(fp8, "%e ", FittedTauAll [i] [bestjtry] [j] );
         fprintf(fp8, "%e\n",FittedAwtAll [i] [bestjtry] [j] );
      }

      fprintf(fp8,"%le ",  AICStatAll  [bestjtry] [j] );
      fprintf(fp8,"%le\n", BICStatAll  [bestjtry] [j] );
   }

   for(j=1; j<=priorKMax; j++)
   {
      if(j==1)
      {
         bestjtry = bestjtryAll [j]; // use bestjtry for one component case
      }
      else
      {
         bestjtry = 1 - bestjtryAll [j];
      }

      for(i=1; i<=j; i++)
      {

         fprintf(fp8, "%e ", FittedMeanAll[i] [bestjtry] [j] );
         fprintf(fp8, "%e ", FittedTauAll [i] [bestjtry] [j] );
         fprintf(fp8, "%e\n",FittedAwtAll [i] [bestjtry] [j] );
      }

      fprintf(fp8,"%le ",  AICStatAll [bestjtry] [j] );
      fprintf(fp8,"%le\n", BICStatAll [bestjtry] [j] );


   }


#if 0 // ============= detailed output if needed ===============

   for( iis = 1; iis<= numoISsamples; iis++)
   {

      fprintf(fp8, "%e %e %e %e %e %e\n",
         AJ4CandProb    [iis], // CandProb
         AJ4PostProb    [iis], // PostProb
         AJ4CompNum     [iis], // k
         AJ4LogCandProb [iis], // logcandprob (with simplex normalisation)
         AJ4LogPostProb [iis], // logpostprob - maxlogpostprob
         AJ4LogRatio    [iis]  // logratio
      );

      // --- Output iis, postprob, k
      fprintf(fp8, "%e %e %e\n",
                C4Cells[iis] [1], /* iis */
                C4Cells[iis] [2], /* Normed Ratio */
                C4Cells[iis] [3] ); /* numcomps */
 
      for(j=1; j<= C4Cells[iis][3]; j++)
      {
         // output mu[j], tau[j], awt[j]
         fprintf(fp8, "%e %e %e\n",
                 C4Cells[iis][j + 3],
                 C4Cells[iis][j + 3 + priorKMax],
                 C4Cells[iis][j + 3 + 2 * priorKMax] );
      }

   }

#endif // ===================================================




   for(i=1; i<=priorKMax; i++)
   {
      fprintf(fp8,"%le %le\n", ProbofK[i], SDProbofK[i]);
   }


}/* out8flag */

// === This next section calculates the weighted average parameter values conditional on k


// === This is the method used by Richardson & Green for estimating par values

   // ======= init counters ========

   for(k=1; k<=priorKMax; k++)
   {
      for(j=1; j<=priorKMax; j++)
      {
         ISmu [j][k] = 0.0;
         IStau[j][k] = 0.0;
         ISawt[j][k] = 0.0;
      }
   }


   istt = 1;
   iend = NumWithKcomps[1];
   for(numcomps=1; numcomps<=priorKMax; numcomps++) // for each k-comp model
   {

      // --- get number of points where given k-comp model is obtained ---
      numpoints = NumWithKcomps[ numcomps];

      ib = 0; // --- ib tallies the # of samples with this k-component
      for( iis = istt; iis<= iend; iis++)
      {
         ib++;
         yorig [ib] = C4Cells[iis][2]; // normed ratio
      }

      for(j=1; j<=numcomps; j++)
      {
         ib = 0; // --- ib tallies the # of the samples with this k-component
         for( iis = istt; iis<= iend; iis++)
         {
            ib++;
            yt = yorig[ib];

            xt = C4Cells[iis][3 + j];
            if(_isnan(xt)!=0) xt = 0.;
            xorig [3*j-2] [ib] = xt;

            x0 = xt * yt;
            ISmu  [j] [numcomps] = ISmu  [j] [numcomps] + x0;

            xt = C4Cells[iis][3 + j + priorKMax];
            if(_isnan(xt)!=0) xt = 0.01;
            xorig [3*j-1] [ib] = xt;

            x0 = xt * yt;
            IStau [j] [numcomps] = IStau [j] [numcomps] + x0;

            xt = C4Cells[iis][3 + j + 2*priorKMax];

            if(_isnan(xt)!=0)
            {
               if(j==numcomps)
               {
                  xt = 1.;
               }
               else
               {
                  xt = 0.;
               }
            }


            xorig [3*j  ] [ib] = xt;
            
            x0 = xt * yt;
            ISawt [j] [numcomps] = ISawt [j] [numcomps] + x0;

         }
      }


//      printf("\n IS estimates of fit for k = %ld components\n", numcomps);
      for(j=1; j<=numcomps; j++)
      {
         ISmu [j][numcomps] = ISmu [j][numcomps]/ProbofK[numcomps];
         IStau[j][numcomps] = IStau[j][numcomps]/ProbofK[numcomps];
         ISawt[j][numcomps] = ISawt[j][numcomps]/ProbofK[numcomps];

         //  ==== diagnostic, these are output to Out8.d anyway ====
         //printf("numcomps=%ld j=%ld pofk=%le: mu=%lf tau=%lf awt=%lf\n", numcomps,
         //           j, ProbofK[numcomps],
         //           ISmu[j][numcomps], IStau[j][numcomps], ISawt[j][numcomps]);

      }
      //getch();


      istt += numpoints;
      iend += NumWithKcomps[numcomps+1];

   }/* numcomps */

// ===== output the IS parameter estimates =====

if(out8flag==1)
{
   for(numcomps=1; numcomps<=priorKMax; numcomps++)
   {
      fprintf(fp8,"%ld\n", numcomps);
      for(j=1; j<=numcomps; j++)
      {
         fprintf(fp8, "%lf %lf %lf\n",
               ISmu[j][numcomps], IStau[j][numcomps], ISawt[j][numcomps]);
      }
   }

   fclose(fp8);
}





// =============== Histo output of k IS samples =====================


akey = 'n'; // --------- diagnostic output ------------

//printf(" See Frequency histograms of the IS samples for each k? (y/b/n)\n");
//akey = getch();


if(akey!='n') // ========= This is just a diagnostic loop ===========

              //  Produces histos of the IS samples for each component k
              //  to give an indication of how well the candidate distn
              //  used in the IS matches the posterior being estimated.
              //  if akey =='n' then this is skipped
{



   /* ----- Prepare the chart info. In this case there are ----- */
   //       priorKMax samples each of size: n = nMCMCsamplesExpected

   // The routine plots the frequency histogram of each sample
   // The values are thus the x-values, these are ordered for each sample


   for(k=1; k<= priorKMax; k++)
   {

       for(j0=1; j0<=50; j0++)
       {
           numpos[j0] = 0;
           numpos1[j0] = 0;

       }

       numpoints = 0;

       // ---- find kth sample into descending order ----

       for(i=1; i<=numoISsamples; i++)
       {
           if(kArray[i]==k)
           {
               if(postProb[i]>0.0)
               {
                  x2 = log(postProb[i])/ log(10); // base 10 exponent of postprob
               }
               else
               {
                  x2 = -500; // effectively postprob treated as zero
               }
               if(x2>-44)
               {
                  numpoints++; 
                  Asort[numpoints] = x2;
                  Bsort[numpoints] = numpoints;         // for sorting
               }


               if(x2 > -1){numpos1[1]++;
               } else if(x2 > -2) {numpos1[2]++;
               } else if(x2 > -3) {numpos1[3]++;
               } else if(x2 > -4) {numpos1[4]++;
               } else if(x2 > -5) {numpos1[5]++;
               } else if(x2 > -6) {numpos1[6]++;
               } else if(x2 > -7) {numpos1[7]++;
               } else if(x2 > -8) {numpos1[8]++;
               } else if(x2 > -9) {numpos1[9]++;
               } else if(x2 > -10) {numpos1[10]++;
               } else if(x2 > -11) {numpos1[11]++;
               } else if(x2 > -12) {numpos1[12]++;
               } else if(x2 > -13) {numpos1[13]++;
               } else if(x2 > -14) {numpos1[14]++;
               } else if(x2 > -15) {numpos1[15]++;
               } else if(x2 > -16) {numpos1[16]++;
               } else if(x2 > -17) {numpos1[17]++;
               } else if(x2 > -18) {numpos1[18]++;
               } else if(x2 > -19) {numpos1[19]++;
               } else if(x2 > -20) {numpos1[20]++;
               } else if(x2 > -21) {numpos1[21]++;
               } else if(x2 > -22) {numpos1[22]++;
               } else if(x2 > -23) {numpos1[23]++;
               } else if(x2 > -24) {numpos1[24]++;
               } else if(x2 > -25) {numpos1[25]++;
               } else if(x2 > -26) {numpos1[26]++;
               } else if(x2 > -27) {numpos1[27]++;
               } else if(x2 > -28) {numpos1[28]++;
               } else if(x2 > -29) {numpos1[29]++;
               } else if(x2 > -30) {numpos1[30]++;
               } else if(x2 > -31) {numpos1[31]++;
               } else if(x2 > -32) {numpos1[32]++;
               } else if(x2 > -33) {numpos1[33]++;
               } else if(x2 > -34) {numpos1[34]++;
               } else if(x2 > -35) {numpos1[35]++;
               } else if(x2 > -36) {numpos1[36]++;
               } else if(x2 > -37) {numpos1[37]++;
               } else if(x2 > -38) {numpos1[38]++;
               } else if(x2 > -39) {numpos1[39]++;
               } else if(x2 > -40) {numpos1[40]++;
               } else if(x2 > -41) {numpos1[41]++;
               } else if(x2 > -42) {numpos1[42]++;
               } else if(x2 > -43) {numpos1[43]++;
               } else if(x2 > -44) {numpos1[44]++;
               }

           }
       }

       printf("\nAt k=%ld \n", k );

       for(j0=1; j0<=44; j0++)
       {
           printf(" numpos1[%ld] = %ld  %d\n",
                     j0, numpos[j0], numpos1[j0] );
       }
       printf("numpoints positive= %ld out of %ld with %ld components\n",
                       numpoints, NumWithKcomps[k], k );
       getch();

       ShellSortintoDescendingTwoCols (1, numpoints, Asort, Bsort); // istt=1 is assumed

       // sortintodescend2(1, numpoints,Asort,Bsort);

       // ======= info passed through to the HistoB routine =======

       numintvls = numpoints/4;  // num frequency intervals
       if(numintvls<200)
       {
          numintvls = 200;
       }
       if(numintvls>500)
       {
          numintvls = 500;
       }

       firsttimeflag = 1;


       // ---- set the number of rows of graphs
       //      and the number in each row, there can be up to 3 in each row

       numGraphs = 1;

       numpoints = numpoints;

      // ------- For each row:

       for(j=1; j<=1; j++)
       {
      
          for( i = 1; i<= numpoints; i++)
          {

          // ==== Enter first column graph x values here 

             xorig [3*j-2] [i] = Asort [i];

             // ==== Enter second column graph x values here, if there are any 
             xorig [3*j-1] [i] = 0;

             // ==== Enter third column graph x values here, if there are any 
             xorig [3*j  ] [i] = 0;
          }
       }


       // enter just one set of y-values, used by all graphs

       for(i=1; i<=numpoints; i++)
       {
          yorig [i] = 1./numpoints; // normed ratio
       }


#if GRAHICSON
       akey =  'y';
       if(akey=='y'|| akey=='b')  // -- output marginal posteriors of params of each component
       {
          // the only difference between this and the original version is that
          // this jumps past the 2nd and 3rd column graphs

          HistoOnlyB(
              xorig,     /* IN: the points in columns */
              yorig,     /* IN: the prob vals in one column */
              numintvls, /* IN: # of histogram cells */
              freq,      /* OUT: cell frequencies */
              numatk,    /* OUT: number of obsns */
              hmax,      /* OUT: max histo height */
              numGraphs, /* IN: # of parameters being plotted */
              numpoints, /* IN: # of points */
              k,         /* IN: # of components */
              ProbofK,   /* IN: prob of k */
              firsttimeflag /* IN: Set to 1 if HistoOnly called only once */
            );


       } // --- akey loop end
#endif


   } // --- end of k-loop

} // outer a-key loop end

// ==== End of section making histogram display of IS samples for each k
//                                                             ============



// ====================== Param posterior probs =============================

#if GRAPHICSON
   printf(" See the frequency histograms of the parameter postprobs? (y/b/n)\n");
   akey = getch();
#else
   printf(" Graphics suppressed. Press key to continue\n");
   getch();
   akey = 'y';
#endif


if(akey!='n') // ===== this loop is carried out if graphics are suppressed 
{

   // ======= Plot out the posteriors of the parameters =======
   numintvls =200;
//   firsttimeflag = 1;

   istt = 1;
   iend = NumWithKcomps[1];

   // === this is a rather crude trimming of the sample range to remove
   // === uninteresting tails where the posterior is near zero
   istt0 = istt + (iend-istt+1)/5;
   iend0 = iend - (iend-istt+1)/5;

   for(numcomps=1; numcomps<=priorKMax; numcomps++) // for each k-comp model
   {

      // --- get number of points where given k-comp model is obtained ---
      numpoints = NumWithKcomps[ numcomps];
      numpoints0 = iend0 - istt0 +1;


      ib = 0; // --- ib tallies the # of samples with this k-component
      for( iis = istt0; iis<= iend0; iis++)
      {
         ib++;
         yorig [ib] = C4Cells[iis][2]; // normed ratio
      }

      for(j=1; j<=numcomps; j++)
      {
         ib = 0; // --- ib tallies the # of the samples with this k-component
         for( iis = istt0; iis<= iend0; iis++)
         {
            ib++;
            yt = yorig[ib];

            xt = C4Cells[iis][3 + j];
            xorig [3*j-2] [ib] = xt;

            xt = C4Cells[iis][3 + j + priorKMax];
            xorig [3*j-1] [ib] = xt;

            xt = C4Cells[iis][3 + j + 2*priorKMax];
            xorig [3*j  ] [ib] = xt;
         }
      }


      istt += numpoints;
      iend += NumWithKcomps[numcomps+1];

      istt0 = istt + (iend-istt+1)/5;
      iend0 = iend - (iend-istt+1)/5;


      numpars = 3*numcomps;

      if(numcomps==1)
      {
         firsttimeflag = 1;
      }
      else if(numcomps==priorKMax)
      {
         firsttimeflag = -1;
      }
      else
      {
         firsttimeflag = 0;
      }


      #if OUTPUTPDF

      if(numcomps==1)
      {

         fopdf = fopen(DataFolderOutPdf,"w");

      }

      #endif

      #if OUT7

      if(numcomps==1)
      {

         fp7 = fopen(DataFolderOut7,"w");

         fprintf(fp7,"%ld\n",priorKMax);
      }

      #endif

      akey =  'y';
      if(akey=='y'|| akey=='b')  // -- output marginal posteriors of params of each component
      {

      HistoOnly(
              xorig,      /* IN: the points in columns */
              yorig,      /* IN: the prob vals in one column */
              numintvls,  /* IN: # of histogram cells */
              freq,       /* OUT: cell frequencies */
              numatk,     /* OUT: number of obsns */
              hmax,       /* OUT: max histo height */
              numpars,    /* IN: # of parameters being plotted */
              numpoints0, /* IN: # of points */
              ProbofK,    /* IN: frequency histo of prob of K */
              SDProbofK,  /* IN: SD */
              fopdf       /* IN/OUT: */
            );
      }


#if OUTPUTPDF

   if(numcomps==priorKMax)
   {
      fclose(fopdf);
   }

#endif

#if OUT7

   if(numcomps==priorKMax)
   {
      fclose(fp7);
   }

#endif



   }/* numcomps */
}


printf("Output in files Out7.d and Out8.d\n");
printf("Press key to finish\n");

getch();


}/*0*/
/* =========== end of main ========== */


/* ==================== determinant ==================== */

double Determinant( long dimension, double matrix[][PSIZE])
{
   long i, j, k;
   double Det;
   double tempFactor;
   static double newMatrix[PSIZE][PSIZE];

   for(i = 1; i<= dimension; i++)
   {
       for(j = 1; j <= dimension; j++)
       {
           newMatrix[i][j] = matrix[i ][j];
       }
   }

   for(k = 1; k < dimension; k++)
   {
       for(i = k + 1; i<= dimension; i++)
       {
           if ( newMatrix[k][ k] == 0) tempFactor = 1;
           else tempFactor = -newMatrix[i][ k] / newMatrix[k][k];
        
           for(j = k + 1; j <= dimension; j++)
           {
              newMatrix[i][ j] = newMatrix[i][ j] + tempFactor * newMatrix[k][ j];
           }
       }
   }

   Det = 1.0;
   for( i = 1; i<= dimension; i++)
   {
       Det = Det * newMatrix[i][ i];
   }

   return Det;

}


double mag(x)
 double x;

 {
 if(x<0.0) return -x;
 else return x;
 }


void detzero(void)
 {
 printf(" determinant is at or near to zero \n");
 }

/*================== order data ===================*/

// ============== this is a crude bubble sort; not used ==================

void sortintoascend( long n, double* a, double* b, double* c )

{/*0*/

double dummy;
int i,j;

for (i=1;i<n;i++)
{/*3*/
   for (j=n-1;j>=i;j--)
   {/*4*/
      if (a[j]>a[j+1])
      {/*5*/

         dummy = a[j];
         a[j] = a[j+1];
         a[j+1] = dummy;

         dummy = b[j];
         b[j] = b[j+1];
         b[j+1] = dummy;

         dummy = c[j];
         c[j] = c[j+1];
         c[j+1] = dummy;

      }/*5*/
   }/*4*/
}/*3*/

}/*0*/



// ================== determinant calc ============================


double matrixDeterminant(double matrix[][PSIZE], long dimension)
{ 

long i,j,k,el;
long jmax,kmax;
long parity;
double max;
double Determinant;
double tempFactor;
double x;

static double newMatrix[PSIZE][PSIZE];


for(i=1; i<=dimension; i++)
{
    for(j=1; j<=dimension; j++)
    {
        newMatrix[i][j] = matrix[i][j];
    }
}

parity = 2*dimension*dimension;
for(i=1;i<=dimension;i++)
{

   // identify largest element of submatrix
   max = -20;
   for(j=i;j<=dimension;j++)
   {
      for(k=i;k<=dimension;k++)
      {
         x = fabs(newMatrix[j][k]);
         if(x > max )
         {
            jmax = j; kmax = k; max = x;
         }
      }
   }

   // swap i and kmax cols; 
   for(el=1;el<=dimension;el++)
   {
      x = newMatrix[el][kmax];
      newMatrix[el][kmax] = newMatrix[el][i];
      newMatrix[el][i] = x;
   }
   // swap i and jmax rowss; 
   for(el=1;el<=dimension;el++)
   {
      x = newMatrix[jmax][el];
      newMatrix[jmax][el] = newMatrix[i][el];
      newMatrix[i][el] = x;
   }

   parity = parity + kmax + jmax;
}

for(k=1; k<dimension; k++)
{
    for(i=k+1; i<=dimension; i++)
    {
        if(newMatrix[k][k] == 0)
        {
            tempFactor = 1;
        }
        else
        {
            tempFactor = -newMatrix[i][k] / newMatrix[k][k];
        }
        for(j=k+1; j<=dimension; j++)
        {
            newMatrix[i][j] = newMatrix[i][j] + tempFactor * newMatrix[k][j];
        }
    }
}

Determinant = 1.0;
for(i=1; i<=dimension; i++)
{
    Determinant = Determinant * newMatrix[i][i];
}

return Determinant;

}// End Function



// ======================= logCandPDF1 ======================

double logCandPDF1 (
        long NumCandiTries[], 
        long iattempt,
        long model,
	double FittedMeanAll[][TWO][MAXNUMCOMPS],
	double FittedTauAll[][TWO][MAXNUMCOMPS],
	long k,
        long jtry,
        double FittedAwtAll[][TWO][MAXNUMCOMPS],

        /* double sampleBeta, */

        double nudof,
	double sampleLsAll[][PSIZE][TWO][MAXNUMCOMPS],
        long bestjtryAll[],
        /* double tauConst, */
	double mu[],
	double tau[],
	double awt[],
        long iis,
        long outnum,
        double stDevT,
        double invsamLsAll[][PSIZE][TWO][MAXNUMCOMPS],
        double mu1[],  // this is an input
        double tau1[], // this is an input
        double awt1[]  // this is an input
	)
{
// ---This function returns the conditional pdf of having sampled the
// ---tau(), mu() and w() that it outputs from the input z()

extern double TVargen(double);


extern double matrixDeterminant(double bmatrix[][PSIZE], long);
extern double gammalog(double, long);

// static double FittedAlpha[MAXNUMCOMPS][PSIZE]; // note these are define
static double FittedAwt[MAXNUMCOMPS][PSIZE]; // with cols swapped
static double FittedMean[MAXNUMCOMPS][PSIZE];
static double FittedTau[MAXNUMCOMPS][PSIZE];

static double sampleL[MAXNUMCOMPS][PSIZE][PSIZE];

static double z[PSIZE];

long i, j, el;
double ySum;
double aSum;
long ifault;

double tempProduct; // ---This variable is det(dw/dy)*det(dy/dxi)
double LogtempProduct;
double LogTProb;
double LogZProb;

// double stDevT; // recalculated here
double squareSum;
double xd;

static double dWbyDY [PSIZE][PSIZE];
static double dWbyDXi [PSIZE][PSIZE];
static double dWbyDZ[PSIZE][PSIZE];

static double x[ PSIZE];
static double y[ PSIZE];

static double alpha[ PSIZE ];
static double ksi[ PSIZE ];
static double bMatrix[ PSIZE ][ PSIZE ];
static double cMatrix[ PSIZE ][ PSIZE ];

FILE * fout2; // diagnostic output
FILE * fout3; // diagnostic output
FILE * fp4; // diagnostic output

long numpars;

// if(iis==outnum) fout3 = fopen("out3.d","w");

// get k component Fitted
for(j=1; j<=k; j++)
{
   FittedMean[k][j] = FittedMeanAll[j][ jtry ][k];
   FittedTau [k][j] = FittedTauAll [j][ jtry ][k];
   FittedAwt [k][j] = FittedAwtAll [j][ jtry ][k];

//   printf("k=%ld: j=%ld mu=%lf, tau=%lf, awt=%lf\n",k, j,
//               FittedMean[k][j],FittedTau[k][j],FittedAwt[k][j]);
}

/* =================== out4.d; not normally output ============== */

// fp4 = fopen("out4.d","w");
numpars = 3*k;
for(i=1; i<=numpars; i++)
{
    for(j=1; j<=numpars; j++)
    {
        sampleL[k][ i][ j]  = sampleLsAll[i][j][ bestjtryAll[k] ][k];
//        fprintf(fp4," %lf", sampleL[k][i][j]);
//        if(j%numpars==0) fprintf(fp4,"\n");
    }
}

// fclose(fp4);
/* ======================================== */

// =============== Calculate z[] array =================
again:
NumCandiTries[k]++; // --- count the total number of attempts so far made for this k. 

//    printf("k= %ld  Numtries = %ld\n", k, NumCandiTries[k]);

if(iattempt==1)
{
   for(i=1; i<3*k; i++)
   {
      // xd = genrand();  // Feb2013 change
      // z[i] = xinvnm( xd );
      z[i] = TVargen( nudof )/stDevT; // this is a standardized t
   }
}
else // iattempt==2.In this version this is not ever used
{
       // --- calc new z's from ivsamLsAll
       for(i=1; i<3*k; i++)
       {
          z[i] = 0.0;
          for(j=1; j<=k; j++)
          {
             z[i] += invsamLsAll[i][j][jtry][k] * (mu1[j] - FittedMean[k][j]);
          }
          for(j=1; j<=k; j++)
          {
             z[i] += invsamLsAll[i][k+j][jtry][k] * (tau1[j] - FittedTau[k][j]);
          }
          for(j=1; j<k; j++)
          {
             z[i] += invsamLsAll[i][2*k+j][jtry][k] * (awt1[j] - FittedAwt[k][j]);
          }
       }
}

for(i=1; i<=k; i++)
{
    x[i] = FittedMean[k][i];
    x[i+k] = FittedTau[k][i];
    x[i+ 2*k] = FittedAwt[k][i];
}

// ==== This is the sample theta in my notes. Last z not used
for(i=1; i<=numpars; i++)
{
    for(j=1; j<numpars; j++)
    {
        x[i] = x[i] + sampleL[k][ i][j] * z[j];
    }
}
/*=========
printf("iattempt = %ld k=%ld\n",iattempt, k);
for(i=1;i<=k;i++)
{
 printf("x[%ld]=%lf\n",i,x[i]);
}
getch();
=========*/


// ===== Added 25 August 2011 ===== for simplified IS version ====
for(i=1; i<=k; i++)
{
    mu[i] = x[i];
    tau[i] = x[i+k];
    awt[i] =x[i+2*k];

    // ---For selected models, transform mu[i] so it is greater than zero
    if(
       model==WMIXTURE ||
       model==GMIXTURE ||
       model==IMIXTURE
      )
    {
       if (mu[i] <= 0.0001) goto again;
    }

    
    if(tau[i]<=0.0) goto again;

    if(awt[i] < 0.0) goto again;
    if(awt[i] > 1.0) goto again;

}



// check order of means ===
//   This is to avoid the so-called exchangeability problem.
//   i.e. need to stay in a single permutation cone to avoid
//   confusion counting permutation images of the same point.


if(k!=1 && iattempt==1)
{
   for(i=1;i<k; i++)
   {
        if(x[i]>x[i+1]) goto again;
   }
}

//printf("OK!\n");
//getch();


// Calculate the pdf of the (non-singular) IS distribution for this replication for the (reduced) Psi-parameter vector.

for(i=1; i<3*k; i++)
{
    for(j=1; j<3*k; j++)
    {
        cMatrix[i][j] =  sampleL[k][ i][j];
    }
}
tempProduct = fabs(matrixDeterminant(cMatrix, 3 * k - 1) );

if(tempProduct<=0)
{
    tempProduct = -15;
}

LogtempProduct = log(tempProduct);

/* ===================
squareSum = 0;
for(i=1; i<numpars; i++) // ===== note there is NO z[numpars]
{
    squareSum += z[i]*z[i];
}

LogZProb = - 0.5 * squareSum - (numpars - 1) * 0.5 * log(2 * PI);
=================== */

#if 1
// =============================
LogTProb = 0.0;
for(i=1; i<numpars; i++) // note there is NO z[numpars]
{

    LogTProb += log( 1. + pow(stDevT * z[i], 2.) /nudof );
}
LogTProb = - 0.5 * (nudof + 1.) *LogTProb;

LogTProb = LogTProb + (numpars - 1.) *
           ( gammalog(0.5 * (nudof + 1.) , ifault)
               - gammalog( 0.5 * nudof, ifault)
               - 0.5 * sqrt(nudof * PI)
		+ stDevT) ;

//  ==================================
#endif


// return  LogZProb - LogtempProduct;

return  LogTProb - LogtempProduct;

}

long CatSample(
double p[],
long dimension)

{
long i;
double invNormalisingFactor;
static double f[PSIZE];
double randomNumber;

// ---- Normalise probabilities ----
   invNormalisingFactor = 0.0;
   for(i=1; i<=dimension; i++)
   {
      invNormalisingFactor = p[i] + invNormalisingFactor;
   }

   f[0] = 0.0;

   for (i=1; i<=dimension; i++)
   {
      p[i] = p[i] / invNormalisingFactor;
      f[i] = f[i - 1] + p[i];
   }

// ---- Sample from distribution ----
   randomNumber = genrand();
//   irancount++;
//   printf("iran=%ld xd=%lf\n",irancount, randomNumber);
//   getch();

   for (i=1; i<=dimension; i++)
   {
      if( randomNumber < f[i] )
      {
         return i;
      }
   }
   return 0.;
}


/*====== inverse cdf of N(0,1) ========*/
double xinvnm(double x)
{
   double a;
   double d,t;
   double xinv;

   if(x <= .5) a = x;
   else        a = 1.0 - x;

   t=sqrt(-2*log(a));
   d=1+t*(1.432788+t*(.189269+t*.001308));
   xinv=t-(2.515517+t*(.802853+t*.010328))/d;
   if ( x <= .5 ) xinv=-xinv;
   return xinv;
}



/* ===== GS generator ===== */
double GSgen(double a)
{
   double b;
   double P;
   double U1, U2;
   double X;

   b = exp(1.0)/(a + exp(1.0) );

step1:
   U1 = genrand();
   P = b*U1;
   if( P > 1. ) goto step2;
   X = pow(P, 1./a);
   U2 = genrand();
   if( U2 <= exp(-X) ) return X;
   else goto step1;

step2:
   X = -log( (b-P)/a );
   U2 = genrand();
   if( U2 <= pow(X, a-1.) ) return X;
   else goto step1;
}

/*=== gamma generator Y with pdf = [x^(a-1)] exp(-x)/G(a)
to get:
   f=(b^-a)[x^(a-1)] exp(-x/b)/G(a) with
simply return with  X = bY when 
  mean(X)=ab, var(X) = abb

( to get (chi-squared with k dof) set a=k/2, b=2 )
========================*/
    
double GBgen(double a)
{
   double A, b, c;
   double U, V, W, Y, Z;
   double t = 4.5;
   double d = 1. + log(4.5);
   if(a<1.0)
   {
      printf("In GBgen a=%lf < 0\n",a);
      goto ExitLabel;
   }

   A=1/sqrt(2*a-1);
   b = a - log(4);
   c = a +1/A;

again:
   U = genrand();
   V = A * log(U/(1.-U)); Y = a*exp(V); Z=U*U*genrand(); W = b + c*V -Y;
   if( W + d - t*Z >= 0.0 ) return Y;
   if( W >= log(Z) ) return Y;
   goto again;

ExitLabel:

return -999999;

}

double GAMgen(double a)
{
   if(a>1.0) return GBgen(a);
   else if(a>0.0) return GSgen(a);
   else
   {
      printf("In GAMgen: a=%lf out of range \n",a);
      goto ExitLabel;
   }

ExitLabel:
return -999999;

}

/* Variate with t-distn with nu degrees of freedom */
double TVargen(double nu)
{
   double chi, x, z;
   z = xinvnm( genrand() );
   chi = 2.0*GAMgen(nu/2.)/nu;
   return z/sqrt(chi);
}

/* ======== INVERSE GAUSSIAN generator ========= */
double raninvgauss(double xmu, double phi)
{
   double Z,T,Y;
   Z = xinvnm( genrand());
   Y = Z*Z;
   T = 1. + ( Y - sqrt(4. * phi * Y + Y * Y )  )/(2. * phi);
   if( genrand() <= 1./(1.+T) ) return xmu*T;
   else return xmu/T;
}


#if 0 //====== order data ====
void sortintoascend2( long m, long n, double* a, double* b)
/* sorts the segment from m to n inclusive */
{/*0*/

double dummy;
int i,j;

for (i=m;i<n;i++)
{/*3*/
   for (j=n-1;j>=i;j--)
   {/*4*/
      if (a[j]>a[j+1])
      {/*5*/

         dummy = a[j];
         a[j] = a[j+1];
         a[j+1] = dummy;

         dummy = b[j];
         b[j] = b[j+1];
         b[j+1] = dummy;

      }/*5*/
   }/*4*/
}/*3*/

}/*0*/

#endif //=============================

/*====== order data ====*/
void sortintodescend2( long m, long n, double* a, double* b)
/* sorts the segment from m to n inclusive */
{/*0*/

double dummy;
int i,j;

for (i=m;i<n;i++)
{/*3*/
   for (j=n-1;j>=i;j--)
   {/*4*/
      if (a[j]<a[j+1])
      {/*5*/

         dummy = a[j];
         a[j] = a[j+1];
         a[j+1] = dummy;

         dummy = b[j];
         b[j] = b[j+1];
         b[j+1] = dummy;

      }/*5*/
   }/*4*/
}/*3*/

}/*0*/


double gaussj(long n, double ao[][PSIZE], double ainv[][PSIZE])
{
   long i,j,k,l,ll,irow,icol;
   static long ipiv[PSIZE], indxr[PSIZE], indxc[PSIZE];
   static double a[PSIZE][PSIZE], b[PSIZE];

   double big,temp,pivinv;

   for(i=1; i<=n; i++)
   for(j=1; j<=n; j++)
   {
      a[i-1][j-1] = ao[i][j]; 
   }

   for(i=0;i<n;i++)ipiv[i]=0;

   for(i=0;i<n;i++)
   {
      big=0;
      for(j=0;j<n;j++)
      {
         if(ipiv[j]!=1)
         {
            for(k=0;k<n;k++)
            {
               if(ipiv[k]==0)
               {
                  if(fabs(a[j][k]) >= big)
                  {
                     big=fabs(a[j][k]);
                     irow=j;
                     icol=k;
                  }
                  else
                  {
                     if(ipiv[k]>1)
                     {
                        printf(" error 1 /n");
                        scanf(" %lf",temp);
                        return 1;
                     }
                  }
               }
            }
         }
      }
      ipiv[icol]+=1;
      if(irow!=icol)
      {
         for (l=0;l<n;l++)
         {
            temp=a[irow][l];
            a[irow][l]=a[icol][l];
            a[icol][l]=temp;
         }
         temp=b[irow];
         b[irow]=b[icol];
         b[icol]=temp;
      }
      indxr[i]=irow;
      indxc[i]=icol;
      if(a[icol][icol]==0)
      {
         printf(" singular matrix %d /n",icol);
         scanf(" %lf",temp);
         return 1;
      } /* ENDIF ON SINGULAR MATRIX */
      pivinv=1/a[icol][icol];
      a[icol][icol]=1;
      for (l=0;l<n;l++)a[icol][l]*=pivinv;
      b[icol]*=pivinv;
      for (ll=0;ll<n;ll++)
      {
         if(ll!=icol)
         {
            temp=a[ll][icol];
            a[ll][icol]=0;
            for(l=0;l<n;l++)a[ll][l]-=a[icol][l]*temp;
            b[ll]-=b[icol]*temp;
         } /* ENDIF */
      }
   } /*   end of main loop over cols to be replaced */

   // printf(" indxr ");
   // for(l=0;l<n;l++)printf(" %d ",indxr[l]);
   // printf("/n indxc ");
   // for(l=0;l<n;l++)printf(" %d ",indxc[l]);

   for (j=0;j<n;j++)
   {
     l=n-1-j;
     if(indxr[l]!=indxc[l])
     {
        for (k=0;k<n;k++)
        {
           temp=a[k][indxr[l]];
           a[k][indxr[l]]=a[k][indxc[l]];
           a[k][indxc[l]]=temp;
        }
     }
  }
  for(i=1; i<=n; i++)
  for(j=1; j<=n; j++)
  {
     ainv[i][j] = a[i-1][j-1]; 
  }

  return 0;
}

/* =================== Histogram Only ========================*/
void HistoOnly(
             double xorig[][BSIZE],  /* IN: the points in columns */
             double yorig[],         /* IN: the postprob in a column */
             long numintvls,              /* IN: # of histogram cells */
             double freq[][MAXNUMINTVLS], /* OUT: cell frequencies */
             long numatk[],               /* OUT: number of obsns */
             double hmax[],               /* OUT: max histo height */
             long numpars,                /* IN: # of parameters being done */
             long numpoints,              /* IN: # of points */
//             double fval[][MAXNUMINTVLS], /* NOT USED: fitted pdf */
//             double parc[][PSIZE],    /* NOT USED: parameters of pdfs */
             double ProbofK[],          /* incidental info */
             double SDProbofK[],        /* incidental info */
             FILE * fopdf                /* output pdf file */

          )

{/*his*/
#if (WATCOM || PC)
static double hist[MAXNUMINTVLS];
static double pk[MAXNUMINTVLS];
static double pdfval[MAXNUMINTVLS];
#endif

char akey;



double histcellw;   /* in original units */
double histcellh;
double histcellwnorm;

long i;
long ialf;
long ik;  /*-subscript of array holding number value of the degree k -*/
long ikderive; /* subscript of array holding derived degree */
long imu;
long isig;
long ixcell;

long j;

static double lbx[40], lby[40];

long nh, nr;
long numcomps;

// long numpdfpts;

double ptot;

static double rtx[40], rty[40];

double x0;
double xa, xaorig;
double xb, xborig;
static double xmax[PSIZE], xmin[PSIZE];
static double xrangeorig[PSIZE];  /* range of xvalues in original units */
double xleft;

double ya, yb;
double yaorig, yborig;
double ybot;
static double ymax[PSIZE];
double ytop;
double ywidth, ywidthm;
double yorigtotal;

for(i=0; i<=numpars+1; i++)
for(j=0; j<MAXNUMINTVLS; j++)
   freq[i][j] = 0.0;

/* ===== there are numcomps rows of graphs ===== */
numcomps = numpars/3;

ywidth = 1.0/(double)numcomps;

/* ======== left graph positions (for the outputfile only) ====== */
for(i=1;i<=numcomps;i++)
{
   lbx[3*i-2] =0.01;                      rtx[3*i-2] = 0.33;
   lby[3*i-2] =1.0-((double)i)*ywidth; rty[3*i-2] = lby[3*i-2] + ywidth;
   
}

/* ======== mid graph positions (for the outputfile only) ====== */
for(i=1;i<=numcomps;i++)
{
   lbx[3*i-1] =.33;                    rtx[3*i-1] = 0.66;
   lby[3*i-1] =1.0-((double)i)*ywidth; rty[3*i-1] = lby[3*i-1] + ywidth;
   
}

/* ======== right graph positions (for the outputfile only) ====== */
for(i=1;i<=numcomps;i++)
{
   lbx[3*i] =0.66;                   rtx[3*i] = .99;
   lby[3*i] =1.0-((double)i)*ywidth; rty[3*i] = lby[3*i] + ywidth;
}

#if GRAPHICSON
   /*---Set screen---*/
   Initgraphics(); /*640x480 16 col*/
#endif

   /*====== HISTOGRAM =====*/

// moved outside subroutine

//   #if OUTPUTPDF
//   if(firsttimeflag==1)
//   {
//      fopdf = fopen("outpdf.d","w");
//   }
//   #endif


  // printf("numpoints =%ld\n numpars= %ld\n",numpoints,numpars);
  // getchgen();

   /* === select number of pdf graph points === */
   // numpdfpts = 200;


#if GRAPHICSON
   setcolgen(WHITE);
   //   clearscreengen();
#endif

   printf("ProbofK[%ld] = %lf SD = %lf\n",
                      numcomps, ProbofK[numcomps], SDProbofK[numcomps]);

   yorigtotal = 0.0;
   for(i=1;i<=numpoints; i++)
   {
      yorigtotal = yorigtotal + yorig[i];
   }
   if(yorigtotal<= 0.0) printf("No observations to plot\n");

   /* ---------------- main calculation ------------------ */
   for(nh=1;nh<=numpars;nh++)  //number of histogram
   {/*nh*/
      /* --- incremental count of number of points --- */
//      if(firsttimeflag==1) numatk[nh]=numpoints;
//      else numatk[nh]+=numpoints;

      numatk[nh]=numpoints;

//      if(firsttimeflag==1) /* find range just once at beginning */
      {
         xmax[nh] = -20; xmin[nh] = 1e20;
         for(i=1;i<=numpoints;i++)
         {
            if(xorig[nh][i]>xmax[nh]) xmax[nh] = xorig[nh][i];
            if(xorig[nh][i]<xmin[nh]) xmin[nh] = xorig[nh][i];
         }
  
         /*----- range -----*/
         xrangeorig[nh] =  xmax[nh] - xmin[nh];
//         printf("xmin=%lf xmax=%lf\n",xmin[nh],xmax[nh]);

         if(xrangeorig[nh] == 0.) /* - if observations form a single spike - */
         {  /* --- then turn this into a rectangular block --- */
//            printf("!!! xmin=%lf xmax=%lf\n",xmin[nh],xmax[nh]);
//            getch();
            xmax[nh] += 0.5;
            xmin[nh] -= 0.5;
            xrangeorig[nh] = 1.0;
         }

         /* --- rescale width --- */
         xmin[nh] = xmin[nh] - xrangeorig[nh]/10.0;
         xmax[nh] = xmax[nh] + xrangeorig[nh]/10.0;
         xrangeorig[nh] =  xmax[nh] - xmin[nh];
      }

//      printf("nh=%2ld xmin=%8.3lf xmax=%8.3lf ",nh,xmin[nh],xmax[nh]);
//      printf("Xrange = %8.3lf\n",xrangeorig[nh]);

      /*----- size of histo cells scaled
              each cell scaled to have unit area in original units
      There are numatk[nh] of them  -----*/
      histcellw = xrangeorig[nh]/(double)numintvls;
      histcellh = 1.0/histcellw; /* cell ht in original units */

      /* ---------- build up frequency counts ------------ */
      for(i=1;i<=numpoints;i++)
      {/*i*/
         /*----- identify and incr the intvl into which the point falls;
                 horiz divided into HRES/histcellw intvls -----*/

         ixcell = 0.5 + (long)( 0.99 * (xorig[nh][i]-xmin[nh]) /histcellw);
         freq[nh][ixcell] += yorig[i]; /*-- frequency is the actual count --*/
                                       /*WAS 1.0 */
      }/*i*/

      /* ------ rescale histobars into pixels ------ 

      on entry each cell has unit height in orig units;
      multiplying by histcellh makes each cell of unit volume, so
      dividing by total num of units makes the total volume unity
      in original units.
      hmax already set at max pdfval on entry at this point
      ------------------------------------------------*/
      
      hmax[nh] = 0.0;

      for(i=1;i<=numintvls;i++)
      {
         hist[i] = freq[nh][i]*histcellh/numatk[nh];
         if(hist[i]>hmax[nh]) hmax[nh] = hist[i];
      }

      /*----- graph dimensions -----*/
      xleft = 1 + ((nh-1)%3)*HWIDTH/3;
      histcellwnorm = (HWIDTH/3)/(double)numintvls;
      ywidth = VHEIGHT/(pow(numpars, 0.66)+1); // scaled so graph sizes do not change too fast
      ywidthm = YCOVER*ywidth;
      ybot = ywidth*( (long)( (nh-1)/3 ) );
      ytop = ybot + ywidthm;
      yborig = 0.0;

      /*----------------------
      now scale to y-pixel value by reducing by hmax
      -----------------------*/
      for(i=1;i<=numintvls;i++)
      {
         hist[i] = ywidthm * hist[i]/hmax[nh] + ybot;
      }
   

#if GRAPHICSON

      /* ------------- draw the histogram --------------- */

      /*--------- plot axes of frequency histogram graph ----------*/
      setcolgen(WHITE);
      movegen(xleft,ytop); drawgen(xleft+HWIDTH/3,ytop);   /*xaxis-top*/
      movegen(xleft,ybot); drawgen(xleft+HWIDTH/3,ybot);   /*xaxis-bot*/
      movegen(xleft,ytop); drawgen(xleft,ybot);         /*yaxis-l*/
      movegen(xleft+HWIDTH/3,ytop); drawgen(xleft+HWIDTH/3,ybot); /*yaxis-r*/

#endif

#if OUTPUTPDF
      fprintf(fopdf,"******* open graph # %2ld.%2ld *******\n",numcomps,nh);
      fprintf(fopdf,"//pos %4.2lf %4.2lf %4.2lf %4.2lf\n",
                         lbx[nh],lby[nh],rtx[nh],rty[nh]);
      fprintf(fopdf,"/ag \n");
      fprintf(fopdf,"/td \"xy\"  ;define columns \n"); /* define format */
      fprintf(fopdf,"//cwt 2  ;change line weight\n"); /* define line weight */
      fprintf(fopdf,"/sm OFF  ;no marker\n"); /* no marker */
      fprintf(fopdf,"/sd OFF  ;contus line\n"); /* solid line  */

      fprintf(fopdf,"%e\t %e\n", xmin[nh], yborig); /*WAS yborig*/
      fprintf(fopdf,"%e\t %e\n", xmax[nh], yborig); /*WAS yborig*/

      fprintf(fopdf,"//nc   ;end of curve \n");
#endif

#if OUT7
      fprintf(fp7,"%2ld %2ld\n",numcomps,nh); //current k and graph number

//  =============    position of graph ignored
//      fprintf(fp7,"%4.2lf %4.2lf %4.2lf %4.2lf\n",
//                         lbx[nh],lby[nh],rtx[nh],rty[nh]); // posn of graph

#endif


#if GRAPHICSON
      setcolgen(GREEN);
#endif



#if OUTPUTPDF
      fprintf(fopdf,"/td \"xy\"  ;define columns \n"); /* define format */
      fprintf(fopdf,"//cwt 2  ;change line weight\n"); /* define line weight */
      fprintf(fopdf,"/sm OFF  ;no marker\n"); /* no marker */
      fprintf(fopdf,"/sd OFF  ;contus line\n"); /* solid line  */

#endif

#if OUT7
      // ---- this is the number of pairs of points in this histo
      fprintf(fp7,"%ld\n", 2*numintvls+2);
#endif

      for(i=1;i<=numintvls;i++)
      {/*i*/
         ixcell = i;
         xa = histcellwnorm*(ixcell-1) + xleft;
         xb = xa + histcellwnorm;
         xaorig = (ixcell /*- 0.5*/ )*histcellw/0.99 + xmin[nh];
         xborig = xaorig + histcellw;
         ya = hist[ixcell];
//         yaorig = freq[nh][i];
         yaorig = hist[i] - ybot;

         yborig = 0.0;

//         fval[nh][i] = xaorig + histcellw/2.0;
                                         /* record orig value of histcell */

#if GRAPHICSON

         if(i==1) movegen(xa,ybot);
         drawgen(xa,ya);
         drawgen(xb,ya);
         if(i==numintvls) drawgen(xb,ybot);
#endif


#if OUTPUTPDF
         if(i==1) fprintf(fopdf,"%e\t %e\n",xaorig, yborig);
         fprintf(fopdf,"%e\t %e\n",xaorig, yaorig);
         fprintf(fopdf,"%e\t %e\n",xborig, yaorig);
         if(i==numintvls) fprintf(fopdf,"%e\t %e\n",xborig, yborig);
#endif

#if OUT7 // --- point pairs of histo ---
         if(i==1) fprintf(fp7,"%e\t %e\n",xaorig, yborig);
         fprintf(fp7,"%e\t %e\n",xaorig, yaorig);
         fprintf(fp7,"%e\t %e\n",xborig, yaorig);
         if(i==numintvls) fprintf(fp7,"%e\t %e\n",xborig, yborig);
#endif


      }/*i*/

#if OUTPUTPDF
      fprintf(fopdf,"//nc   ;end of curve \n ");
#endif

   /*------------ end of histogram plot -----------*/

   }/*nh*/


#if GRAPHICSON

   getch(); // ----- pause to see plot

   /* ---- close the graphics ----- */
   Closegraphics();

#endif

// moved outside subroutine
// #if OUTPUTPDF
//if(firsttimeflag==-1)
//{
//   fclose(fopdf);
//}
//#endif

}/*his*/
/*======================= end of HistoOnly =========================*/




#if GRAPHICSON

/* =================== Histogram Only B ========================*/
// Only two lines different picked out by //######## lines
void HistoOnlyB(
             double xorig[][BSIZE],  /* IN: the points in columns */
             double yorig[],         /* IN: the postprob in a column */
             long numintvls,              /* IN: # of histogram cells */
             double freq[][MAXNUMINTVLS], /* OUT: cell frequencies */
             long numatk[],               /* OUT: number of obsns */
             double hmax[],               /* OUT: max histo height */
             long numpars,                /* IN: # of parameters being done */
             long numpoints,              /* IN: # of points */
//             double fval[][MAXNUMINTVLS], /* NOT USED: fitted pdf */
//             double parc[][PSIZE],        /* NOT USED: parameters of pdfs */
             long k,             /* number of components being considered */
             double ProbofK[],              /* incidental info */
             long firsttimeflag /* IN: Set to 1 if HistoOnly called only once */

          )

{/*his*/
#if (WATCOM || PC)
static double hist[MAXNUMINTVLS];
static double pk[MAXNUMINTVLS];
static double pdfval[MAXNUMINTVLS];
#endif

char akey;

FILE * fopdf;

double histcellw;   /* in original units */
double histcellh;
double histcellwnorm;

long i;
long ialf;
long ik;  /*-subscript of array holding number value of the degree k -*/
long ikderive; /* subscript of array holding derived degree */
long imu;
long isig;
long ixcell;

long j;

static double lbx[40], lby[40];

long nh, nr;
long numcomps;

// long numpdfpts;

double ptot;

static double rtx[40], rty[40];

double x0;
double xa, xaorig;
double xb, xborig;
static double xmax[PSIZE], xmin[PSIZE];
static double xrangeorig[PSIZE];  /* range of xvalues in original units */
double xleft;

double ya, yb;
double yaorig, yborig;
double ybot;
double ytop;
double ywidth, ywidthm;
double yorigtotal;

for(i=0; i<=numpars+1; i++)
for(j=0; j<MAXNUMINTVLS; j++)
   freq[i][j] = 0.0;

/* ===== there are numcomps rows of graphs ===== */

numcomps = numpars/3;  //###################

ywidth = 1.0/(double)numcomps;

/* ======== left graph positions (for the outputfile only) ====== */
for(i=1;i<=numcomps;i++)
{
   lbx[3*i-2] =0.01;                      rtx[3*i-2] = 0.33;
   lby[3*i-2] =1.0-((double)i)*ywidth; rty[3*i-2] = lby[3*i-2] + ywidth;
   
}

/* ======== mid graph positions (for the outputfile only) ====== */
for(i=1;i<=numcomps;i++)
{
   lbx[3*i-1] =.33;                    rtx[3*i-1] = 0.66;
   lby[3*i-1] =1.0-((double)i)*ywidth; rty[3*i-1] = lby[3*i-1] + ywidth;
   
}

/* ======== right graph positions (for the outputfile only) ====== */
for(i=1;i<=numcomps;i++)
{
   lbx[3*i] =0.66;                   rtx[3*i] = .99;
   lby[3*i] =1.0-((double)i)*ywidth; rty[3*i] = lby[3*i] + ywidth;
}

   /*---Set screen---*/
   Initgraphics(); /*640x480 16 col*/

   /*====== HISTOGRAM =====*/
   #if OUTPUTPDF
   fopdf = fopen(DataFolderOutPdf,"w");
   #endif


  // printf("numpoints =%ld\n numpars= %ld\n",numpoints,numpars);
  // getchgen();

   /* === select number of pdf graph points === */
   // numpdfpts = 200;
   setcolgen(WHITE);
   //   clearscreengen();

   printf("ProbofK[%ld] = %lf\n", k, ProbofK[k]);

   yorigtotal = 0.0;
   for(i=1;i<=numpoints; i++)
   {
      yorigtotal = yorigtotal + yorig[i];
   }
   if(yorigtotal<= 0.0) printf("No observations to plot\n");

   /* ---------------- main calculation ------------------ */
   for(nh=1;nh<=numpars;nh=nh+3)  //########################
   {/*nh*/
      /* --- incremental count of number of points --- */
      if(firsttimeflag==1) numatk[nh]=numpoints;
      else numatk[nh]+=numpoints;

      if(firsttimeflag==1) /* find range just once at beginning */
      {
         xmax[nh] = -1e20; xmin[nh] = 1e20;
         for(i=1;i<=numpoints;i++)
         {
            if(xorig[nh][i]>xmax[nh]) xmax[nh] = xorig[nh][i];
            if(xorig[nh][i]<xmin[nh]) xmin[nh] = xorig[nh][i];
         }
  
         /*----- range -----*/
         xrangeorig[nh] =  xmax[nh] - xmin[nh];
//         printf("xmin=%lf xmax=%lf\n",xmin[nh],xmax[nh]);

         if(xrangeorig[nh] == 0.) /* - if observations form a single spike - */
         {  /* --- then turn this into a rectangular block --- */
//            printf("!!! xmin=%lf xmax=%lf\n",xmin[nh],xmax[nh]);
//            getch();
            xmax[nh] += 0.5;
            xmin[nh] -= 0.5;
            xrangeorig[nh] = 1.0;
         }

         /* --- rescale width --- */
         xmin[nh] = xmin[nh] - xrangeorig[nh]/10.0;
         xmax[nh] = xmax[nh] + xrangeorig[nh]/10.0;
         xrangeorig[nh] =  xmax[nh] - xmin[nh];
      }

//      printf("nh=%2ld xmin=%8.3lf xmax=%8.3lf ",nh,xmin[nh],xmax[nh]);
//      printf("Xrange = %8.3lf\n",xrangeorig[nh]);

      /*----- size of histo cells scaled
              each cell scaled to have unit area in original units
      There are numatk[nh] of them  -----*/
      histcellw = xrangeorig[nh]/(double)numintvls;
      histcellh = 1.0/histcellw; /* cell ht in original units */

      /* ---------- build up frequency counts ------------ */
      for(i=1;i<=numpoints;i++)
      {/*i*/
         /*----- identify and incr the intvl into which the point falls;
                 horiz divided into HRES/histcellw intvls -----*/

         ixcell = 0.5 + (long)( 0.99 * (xorig[nh][i]-xmin[nh]) /histcellw);
         freq[nh][ixcell] += yorig[i]; /*-- frequency is the actual count --*/
                                       /*WAS 1.0 */
      }/*i*/

      /* ------ rescale histobars into pixels ------ 

      on entry each cell has unit height in orig units;
      multiplying by histcellh makes each cell of unit volume, so
      dividing by total num of units makes the total volume unity
      in original units.
      hmax already set at max pdfval on entry at this point
      ------------------------------------------------*/
      
      hmax[nh] = 0.0;

      for(i=1;i<=numintvls;i++)
      {
         hist[i] = freq[nh][i]*histcellh/numatk[nh];
         if(hist[i]>hmax[nh]) hmax[nh] = hist[i];
      }

      /*----- graph dimensions -----*/
      xleft = 1 + ((nh-1)%3)*HWIDTH/3;
      histcellwnorm = (HWIDTH/3)/(double)numintvls;
      ywidth = VHEIGHT/(pow(numpars, 0.66)+1); // scaled so graph sizes do not change too fast
      ywidthm = YCOVER*ywidth;
      ybot = ywidth*( (long)( (nh-1)/3 ) );
      ytop = ybot + ywidthm;
      yborig = 0.0;

      /*----------------------
      now scale to y-pixel value by reducing by hmax
      -----------------------*/
      for(i=1;i<=numintvls;i++)
      {
         hist[i] = ywidthm * hist[i]/hmax[nh] + ybot;
      }
   
      /* ------------- draw the histogram --------------- */

      /*--------- plot axes of frequency histogram graph ----------*/
      setcolgen(WHITE);
      movegen(xleft,ytop); drawgen(xleft+HWIDTH/3,ytop);   /*xaxis-top*/
      movegen(xleft,ybot); drawgen(xleft+HWIDTH/3,ybot);   /*xaxis-bot*/
      movegen(xleft,ytop); drawgen(xleft,ybot);         /*yaxis-l*/
      movegen(xleft+HWIDTH/3,ytop); drawgen(xleft+HWIDTH/3,ybot); /*yaxis-r*/

#if OUTPUTPDF
      fprintf(fopdf,"******* open graph # %2ld *******\n",nh);
      fprintf(fopdf,"//pos %4.2lf %4.2lf %4.2lf %4.2lf\n",
                         lbx[nh],lby[nh],rtx[nh],rty[nh]);
      fprintf(fopdf,"/ag \n");
      fprintf(fopdf,"/td \"xy\"  ;define columns \n "); /* define format */
#if 1
      fprintf(fopdf,"%e\t %e\n", xmin[nh], yborig); /*WAS yborig*/
      fprintf(fopdf,"%e\t %e\n", xmax[nh], yborig); /*WAS yborig*/
#endif
      fprintf(fopdf,"//nc   ;end of curve \n ");
#endif

      setcolgen(GREEN);
#if OUTPUTPDF
      fprintf(fopdf,"/td \"xy\"  ;define columns \n "); /* define format */
#endif
      for(i=1;i<=numintvls;i++)
      {/*i*/
         ixcell = i;
         xa = histcellwnorm*(ixcell-1) + xleft;
         xb = xa + histcellwnorm;
         xaorig = (ixcell /*- 0.5*/ )*histcellw/0.99 + xmin[nh];
         xborig = xaorig + histcellw;
         ya = hist[ixcell];
//         yaorig = freq[nh][i];
         yaorig = hist[i] - ybot;

         yborig = 0.0;

//         fval[nh][i] = xaorig + histcellw/2.0;
                                         /* record orig value of histcell */

         if(i==1) movegen(xa,ybot);
         drawgen(xa,ya);
         drawgen(xb,ya);
         if(i==numintvls) drawgen(xb,ybot);

#if OUTPUTPDF
         if(i==1) fprintf(fopdf,"%e\t %e\n",xaorig, yborig);
         fprintf(fopdf,"%e\t %e\n",xaorig, yaorig);
         fprintf(fopdf,"%e\t %e\n",xborig, yaorig);
         if(i==numintvls) fprintf(fopdf,"%e\t %e\n",xborig, yborig);
#endif

      }/*i*/

#if OUTPUTPDF
      fprintf(fopdf,"//nc   ;end of curve \n ");
#endif
   /*------------ end of histogram plot -----------*/
   }/*nh*/

   getch(); // ----- pause to see plot
   /* ---- close the graphics ----- */

   Closegraphics();

#if OUTPUTPDF
   fclose(fopdf);
#endif

}/*his*/
/*======================= end of HistoOnlyB =========================*/

#endif


void sortintoascend1( long n, double* a)

{/*0*/

double dummy;
int i,j;

for (i=1;i<n;i++)
{/*3*/
   for (j=n-1;j>=i;j--)
   {/*4*/
      if (a[j]>a[j+1])
      {/*5*/

         dummy = a[j];
         a[j] = a[j+1];
         a[j+1] = dummy;

      }/*5*/
   }/*4*/
}/*3*/

}/*0*/

void sortintoascend2( long n, double* a, double* b)

{/*0*/

double dummy;
int i,j;

for (i=1;i<n;i++)
{/*3*/
   for (j=n-1;j>=i;j--)
   {/*4*/
      if (a[j]>a[j+1])
      {/*5*/

         dummy = a[j];
         a[j] = a[j+1];
         a[j+1] = dummy;

         dummy = b[j];
         b[j] = b[j+1];
         b[j+1] = dummy;


      }/*5*/
   }/*4*/
}/*3*/

}/*0*/

void sortintoascend3( long n, double* a, double* b, double* c )

{/*0*/

double dummy;
int i,j;

for (i=1;i<n;i++)
{/*3*/
   for (j=n-1;j>=i;j--)
   {/*4*/
      if (a[j]>a[j+1])
      {/*5*/

         dummy = a[j];
         a[j] = a[j+1];
         a[j+1] = dummy;

         dummy = b[j];
         b[j] = b[j+1];
         b[j+1] = dummy;

         dummy = c[j];
         c[j] = c[j+1];
         c[j+1] = dummy;

      }/*5*/
   }/*4*/
}/*3*/

}/*0*/



// ---------- ShellSort -----------

void ShellSortintoAscendingOneCol(
        long NumRows,
        double* X
        )
{
    //  sorts into ascending order

    long Gap;
    long DoneFlag;
    long Icount;
    long Index;
    double Swap;
    long J;
    
    Icount = 0;

    // -- Gap is half the records

    Gap = (long)(NumRows / 2);

    while( Gap >= 1)
    {
        do
        {
            DoneFlag = 1;
            for( Index = 1; Index <= NumRows - Gap; Index++)
            {
                // -- Compare X 1st 1/2 to 2nd 1/2
                if( X [Index] > X [Index + Gap] )
                {
                    // -- Swap Values if 1st > 2nd

                    Swap = X [Index];
                    X [Index] = X [Index + Gap];
                    X [Index + Gap] = Swap;
                    
                    Icount = Icount + 1;
                    DoneFlag = 0;
                }
            }
        }while ( DoneFlag == 0);
        Gap = (long)(Gap / 2);
    }
}                                                    

void ShellSortintoAscendingTwoCols(
        long istt, long iend,
        double *X, double *Y
        )
{
    //  sorts first col into ascending order, keeping other cols in step

    long Gap;
    long DoneFlag;
    long Icount;
    long Index;
    double Swap;
    long J;
    long iBase;
    long NumRows;

    Icount = 0;

    NumRows = iend - istt + 1;

    iBase = istt - 1;

    // -- Gap is half the records

    Gap = (long)(NumRows / 2);

    while( Gap >= 1)
    {
        do
        {
            DoneFlag = 1;
            for( Index = iBase + 1; Index <= iBase + NumRows - Gap; Index++)
            {
                // -- Compare X 1st 1/2 to 2nd 1/2
                if( X [Index] > X [Index + Gap] )
                {
                    // -- Swap Values if 1st > 2nd

                    Swap = X [Index];
                    X [Index] = X [Index + Gap];
                    X [Index + Gap] = Swap;
                    
                    Swap = Y [Index];
                    Y [Index] = Y [Index + Gap];
                    Y [Index + Gap] = Swap;

                    Icount = Icount + 1;
                    DoneFlag = 0;
                }
            }
        }while ( DoneFlag == 0);
        Gap = (long)(Gap / 2);
    }
}                                                    



void ShellSortintoAscendingThreeCols(
        long NumRows,
        double *X, double *Y, double *Z
        )
{
    //  sorts first col into ascending order, keeping other cols in step

    long Gap;
    long DoneFlag;
    long Icount;
    long Index;
    double Swap;
    long J;
    
    Icount = 0;

    // -- Gap is half the records

    Gap = (long)(NumRows / 2);

    while( Gap >= 1)
    {
        do
        {
            DoneFlag = 1;
            for( Index = 1; Index <= NumRows - Gap; Index++)
            {
                // -- Compare X 1st 1/2 to 2nd 1/2
                if( X [Index] > X [Index + Gap] )
                {
                    // -- Swap Values if 1st > 2nd

                    Swap = X [Index];
                    X [Index] = X [Index + Gap];
                    X [Index + Gap] = Swap;
                    
                    Swap = Y [Index];
                    Y [Index] = Y [Index + Gap];
                    Y [Index + Gap] = Swap;

                    Swap = Z [Index];
                    Z [Index] = Z [Index + Gap];
                    Z [Index + Gap] = Swap;

                    Icount = Icount + 1;
                    DoneFlag = 0;
                }
            }
        }while ( DoneFlag == 0);
        Gap = (long)(Gap / 2);
    }
}                                                    


void ShellSortintoDescendingTwoCols(
        long istt, long iend,
        double *X, double *Y
        )
{
    //  sorts first col into ascending order, keeping other cols in step

    long Gap;
    long DoneFlag;
    long Icount;
    long Index;
    double Swap;
    long J;
    long iBase;
    long NumRows;

    Icount = 0;

    NumRows = iend - istt + 1;

    iBase = istt - 1;

    // -- Gap is half the records

    Gap = (long)(NumRows / 2);

    while( Gap >= 1)
    {
        do
        {
            DoneFlag = 1;
            for( Index = iBase + 1; Index <= iBase + NumRows - Gap; Index++)
            {
                // -- Compare X 1st 1/2 to 2nd 1/2
                if( X [Index] < X [Index + Gap] )
                {
                    // -- Swap Values if 1st > 2nd

                    Swap = X [Index];
                    X [Index] = X [Index + Gap];
                    X [Index + Gap] = Swap;
                    
                    Swap = Y [Index];
                    Y [Index] = Y [Index + Gap];
                    Y [Index + Gap] = Swap;

                    Icount = Icount + 1;
                    DoneFlag = 0;
                }
            }
        }while ( DoneFlag == 0);
        Gap = (long)(Gap / 2);
    }
}                                                    
